import json
from typing import Mapping, Optional
from unittest import TestCase

from pcs import settings
from pcs.common import file_type_codes, reports
from pcs.common.communication.const import COM_STATUS_SUCCESS
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.communication.types import CommunicationResultStatus
from pcs.common.file_type_codes import PCS_KNOWN_HOSTS
from pcs.common.host import PcsKnownHost
from pcs.common.interface.dto import to_dict
from pcs.common.node_communicator import RequestTarget
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.host.config.types import KnownHosts
from pcs.lib.pcs_cfgsync import save_sync

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


def fixture_communication_result_string(
    status: CommunicationResultStatus = COM_STATUS_SUCCESS,
    status_msg: Optional[str] = None,
    report_list: Optional[reports.dto.ReportItemDto] = None,
    data="",
) -> str:
    return json.dumps(
        to_dict(
            InternalCommunicationResultDto(
                status=status,
                status_msg=status_msg,
                report_list=report_list or [],
                data=data,
            )
        )
    )


def fixture_known_hosts_file_content(
    data_version: int = 1,
    known_hosts: Optional[Mapping[str, PcsKnownHost]] = None,
) -> str:
    return KnownHostsExporter.export(
        KnownHosts(
            format_version=1,
            data_version=data_version,
            known_hosts=known_hosts or {},
        )
    ).decode("utf-8")


def fixture_known_hosts_facade(
    data_version: int = 1,
    known_hosts: Optional[Mapping[str, PcsKnownHost]] = None,
) -> KnownHostsFacade:
    return KnownHostsFacade(
        KnownHosts(
            format_version=1,
            data_version=data_version,
            known_hosts=known_hosts or {},
        )
    )


class FixtureFetchNewestFileMixin:
    def fixture_fetch_newest_file(
        self,
        local_file_content,
        remote_file_content,
        call_name_suffix="",
        cluster_name="cluster",
    ):
        self.config.http.place_multinode_call(
            f"fetch.get_configs{call_name_suffix}",
            node_labels=["node1", "node2"],
            output=fixture_communication_result_string(
                data=SyncConfigsDto(
                    cluster_name=cluster_name,
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: remote_file_content
                    },
                )
            ),
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": cluster_name}),
        )
        # the lib.pcs_cfgsync.fetcher.ConfigFetcher reads the file to decide
        # if the files from cluster are newer than the local file
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name=f"fetch.raw_file.exists{call_name_suffix}",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=local_file_content,
            name=f"fetch.raw_file.read{call_name_suffix}",
        )


class SaveSyncNewVersion(TestCase, FixtureFetchNewestFileMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.local_file_version = 1
        self.local_file = fixture_known_hosts_file_content(1)
        self.local_known_hosts_facade = KnownHostsFacade(
            KnownHosts(
                format_version=1,
                data_version=self.local_file_version,
                known_hosts={},
            )
        )

    def test_success_first_try(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    self.local_file_version + 1
                )
            },
            node_labels=["node1", "node2"],
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_version(
            PCS_KNOWN_HOSTS,
            self.local_known_hosts_facade,
            "cluster",
            [RequestTarget("node1"), RequestTarget("node2")],
            env.get_node_communicator(),
            env.report_processor,
            False,
        )
        self.assertFalse(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
            ]
        )

    def test_conflict_no_fetch(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    self.local_file_version + 1
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {
                    "label": "node2",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "error"},
                        }
                    ),
                },
                {"label": "node3"},
            ],
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_version(
            PCS_KNOWN_HOSTS,
            self.local_known_hosts_facade,
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
                RequestTarget("node3"),
            ],
            env.get_node_communicator(),
            env.report_processor,
            fetch_on_conflict=False,
        )
        self.assertTrue(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2", "node3"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_SAVE_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
            ]
        )

    def test_reject_not_error(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    self.local_file_version + 1
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {
                    "label": "node2",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "error"},
                        }
                    ),
                },
                {"label": "node3"},
            ],
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_version(
            PCS_KNOWN_HOSTS,
            self.local_known_hosts_facade,
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
                RequestTarget("node3"),
            ],
            env.get_node_communicator(),
            env.report_processor,
            fetch_on_conflict=False,
            reject_is_error=False,
        )
        self.assertTrue(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2", "node3"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_SAVE_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
            ]
        )

    def test_conflict_fetch_success(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    self.local_file_version + 1
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
        )
        remote_file = fixture_known_hosts_file_content(42)
        self.fixture_fetch_newest_file(self.local_file, remote_file)

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_version(
            PCS_KNOWN_HOSTS,
            self.local_known_hosts_facade,
            "cluster",
            [RequestTarget("node1"), RequestTarget("node2")],
            env.get_node_communicator(),
            env.report_processor,
            fetch_on_conflict=True,
        )
        self.assertTrue(conflict_detected)
        self.assertEqual(
            new_file.config, KnownHosts(1, data_version=42, known_hosts={})
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
            ]
        )


class SaveSyncNewKnownHosts(TestCase, FixtureFetchNewestFileMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success_first_try(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, {"NODE": PcsKnownHost("NODE", "TOKEN", [])}
                )
            },
            node_labels=["node1", "node2"],
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            fixture_known_hosts_facade(1),
            [PcsKnownHost("NODE", "TOKEN", [])],
            [],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertFalse(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
            ]
        )

    def test_success_first_try_removing_tokens(self):
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, {"NODE-B": PcsKnownHost("NODE-B", "TOKEN", [])}
                )
            },
            node_labels=["node1", "node2"],
        )

        local_file = fixture_known_hosts_facade(
            1,
            {
                "NODE-A": PcsKnownHost("NODE-A", "TOKEN", []),
                "NODE-B": PcsKnownHost("NODE-B", "TOKEN", []),
            },
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            local_file,
            [],
            ["NODE-A"],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertFalse(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
            ]
        )

    def test_conflict_success_after_merge(self):
        local_file_version = 1
        remote_file_version = 42
        local_tokens = {"LOCAL": PcsKnownHost("LOCAL", "LOCAL", [])}
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}
        new_tokens = {"NEW": PcsKnownHost("NEW", "NEW", [])}

        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, local_tokens | new_tokens
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        local_file = fixture_known_hosts_file_content(
            local_file_version, local_tokens
        )
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.fixture_fetch_newest_file(local_file, remote_file)
        # sending the merged tokens
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    remote_file_version + 1,
                    local_tokens | new_tokens | remote_tokens,
                )
            },
            node_labels=["node1", "node2"],
            name="set_configs.2",
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            fixture_known_hosts_facade(local_file_version, local_tokens),
            list(new_tokens.values()),
            [],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertFalse(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
            ]
        )

    def test_conflict_even_more_conflict(self):
        local_file_version = 1
        local_tokens = {"LOCAL": PcsKnownHost("LOCAL", "LOCAL", [])}

        remote_file_version = 42
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}

        new_tokens = {"NEW": PcsKnownHost("NEW", "NEW", [])}

        even_more_new_remote_file_version = 69
        even_more_new_remote_tokens = {"WHAT": PcsKnownHost("WHAT", "WHAT", [])}

        local_file = fixture_known_hosts_file_content(
            local_file_version, local_tokens
        )
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, local_tokens | new_tokens
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        # fetching the newest config from cluster
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file, remote_file, call_name_suffix="1"
        )

        # sending the merged tokens
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    remote_file_version + 1,
                    local_tokens | new_tokens | remote_tokens,
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.2",
        )

        # fetching the even more newest config from cluster
        even_more_new_remote_file = fixture_known_hosts_file_content(
            even_more_new_remote_file_version, even_more_new_remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file, even_more_new_remote_file, call_name_suffix="2"
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            fixture_known_hosts_facade(local_file_version, local_tokens),
            list(new_tokens.values()),
            [],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertTrue(conflict_detected)
        self.assertEqual(
            new_file.config,
            KnownHosts(
                format_version=1,
                data_version=even_more_new_remote_file_version,
                known_hosts=even_more_new_remote_tokens,
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
            ]
        )

    def test_conflict_success_after_merge_remove_tokens(self):
        local_file_version = 1
        remote_file_version = 42
        local_tokens = {
            "LOCAL-A": PcsKnownHost("LOCAL-A", "LOCAL-A", []),
            "LOCAL-B": PcsKnownHost("LOCAL-B", "LOCAL-B", []),
        }
        local_tokens_after_remove = {
            "LOCAL-B": PcsKnownHost("LOCAL-B", "LOCAL-B", []),
        }
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}

        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, local_tokens_after_remove
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        local_file = fixture_known_hosts_file_content(
            local_file_version, local_tokens
        )
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.fixture_fetch_newest_file(local_file, remote_file)

        # sending the merged tokens
        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    remote_file_version + 1,
                    local_tokens_after_remove | remote_tokens,
                )
            },
            node_labels=["node1", "node2"],
            name="set_configs.2",
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            fixture_known_hosts_facade(local_file_version, local_tokens),
            [],
            ["LOCAL-A"],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertFalse(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
            ]
        )

    def test_conflict_error_reading_file_in_fetch(self):
        local_file_version = 1
        remote_file_version = 42
        local_tokens = {"LOCAL": PcsKnownHost("LOCAL", "LOCAL", [])}
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}
        new_tokens = {"NEW": PcsKnownHost("NEW", "NEW", [])}

        self.config.http.pcs_cfgsync.set_configs(
            "cluster",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, local_tokens | new_tokens
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        # fetching the newest config from cluster
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.config.http.place_multinode_call(
            node_labels=["node1", "node2"],
            output=fixture_communication_result_string(
                data=SyncConfigsDto(
                    cluster_name="cluster",
                    configs={file_type_codes.PCS_KNOWN_HOSTS: remote_file},
                )
            ),
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": "cluster"}),
            name="get_configs",
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            exception_msg="error reading during fetch",
        )

        env = self.env_assist.get_env()
        conflict_detected, new_file = save_sync.save_sync_new_known_hosts(
            fixture_known_hosts_facade(local_file_version, local_tokens),
            list(new_tokens.values()),
            [],
            "cluster",
            [
                RequestTarget("node1"),
                RequestTarget("node2"),
            ],
            env.get_node_communicator(),
            env.report_processor,
        )
        self.assertTrue(conflict_detected)
        self.assertIsNone(new_file)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="read",
                    reason="error reading during fetch",
                    file_path=settings.pcsd_known_hosts_location,
                ),
            ]
        )



from dataclasses import dataclass
from typing import Optional

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    FUTURE_OPTION,
    ArgsByKeywords,
    Argv,
    InputModifiers,
    KeyValueParser,
    group_by_keywords,
)
from pcs.cli.reports.output import deprecation_warning


@dataclass(frozen=True)
class PrimitiveOptions:
    instance_attrs: dict[str, str]
    meta_attrs: dict[str, str]
    operations: list[dict[str, str]]


@dataclass(frozen=True)
class CloneOptions:
    clone_id: Optional[str]
    meta_attrs: dict[str, str]


@dataclass(frozen=True)
class GroupOptions:
    group_id: str
    after_resource: Optional[str]
    before_resource: Optional[str]


@dataclass(frozen=True)
class ComplexResourceOptions:
    primitive: PrimitiveOptions
    group: Optional[GroupOptions]
    clone: Optional[CloneOptions]
    promotable: Optional[CloneOptions]
    bundle_id: Optional[str]


@dataclass(frozen=True)
class BundleCreateOptions:
    container_type: str
    container: dict[str, str]
    network: dict[str, str]
    port_map: list[dict[str, str]]
    storage_map: list[dict[str, str]]
    meta_attrs: dict[str, str]


@dataclass(frozen=True)
class BundleUpdateOptions:
    container: dict[str, str]
    network: dict[str, str]
    port_map_add: list[dict[str, str]]
    port_map_remove: list[str]
    storage_map_add: list[dict[str, str]]
    storage_map_remove: list[str]
    meta_attrs: dict[str, str]


@dataclass(frozen=True)
class AddRemoveOptions:
    add: list[dict[str, str]]
    remove: list[str]


def parse_primitive(arg_list: Argv) -> PrimitiveOptions:
    groups = group_by_keywords(
        arg_list, {"op", "meta"}, implicit_first_keyword="instance"
    )

    return PrimitiveOptions(
        instance_attrs=KeyValueParser(
            groups.get_args_flat("instance")
        ).get_unique(),
        meta_attrs=KeyValueParser(groups.get_args_flat("meta")).get_unique(),
        operations=[
            KeyValueParser(op).get_unique()
            for op in build_operations(groups.get_args_groups("op"))
        ],
    )


def parse_clone(arg_list: Argv, promotable: bool = False) -> CloneOptions:
    clone_id = None
    allowed_keywords = {"op", "meta"}
    if (
        arg_list
        and arg_list[0] not in allowed_keywords
        and "=" not in arg_list[0]
    ):
        clone_id = arg_list.pop(0)
    groups = group_by_keywords(
        arg_list, allowed_keywords, implicit_first_keyword="options"
    )

    if groups.has_keyword("op"):
        raise CmdLineInputError(
            "op settings must be changed on base resource, not the clone",
        )
    if groups.has_keyword("options"):
        # deprecated since 0.11.6
        deprecation_warning(
            "configuring meta attributes without specifying the 'meta' keyword "
            "is deprecated and will be removed in a future release"
        )

    meta = KeyValueParser(
        groups.get_args_flat("options") + groups.get_args_flat("meta")
    ).get_unique()
    if promotable:
        if "promotable" in meta:
            raise CmdLineInputError(
                "you cannot specify both promotable option and promotable "
                "keyword"
            )
        meta["promotable"] = "true"
    return CloneOptions(clone_id=clone_id, meta_attrs=meta)


def parse_create_new(arg_list: Argv) -> ComplexResourceOptions:  # noqa: PLR0912
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    top_groups = group_by_keywords(
        arg_list,
        {"clone", "promotable", "bundle", "group"},
        implicit_first_keyword="primitive",
    )
    top_groups.ensure_unique_keywords()

    primitive_groups = group_by_keywords(
        top_groups.get_args_flat("primitive"),
        {"op", "meta"},
        implicit_first_keyword="instance",
    )
    primitive_options = PrimitiveOptions(
        instance_attrs=KeyValueParser(
            primitive_groups.get_args_flat("instance")
        ).get_unique(),
        meta_attrs=KeyValueParser(
            primitive_groups.get_args_flat("meta")
        ).get_unique(),
        operations=[
            KeyValueParser(op).get_unique()
            for op in build_operations(primitive_groups.get_args_groups("op"))
        ],
    )

    group_options = None
    if top_groups.has_keyword("group"):
        group_groups = group_by_keywords(
            top_groups.get_args_flat("group"),
            {"before", "after", "op", "meta"},
            implicit_first_keyword="group_id",
        )
        if group_groups.has_keyword("meta"):
            raise CmdLineInputError(
                "meta options must be defined on the base resource, "
                "not the group"
            )
        if group_groups.has_keyword("op"):
            raise CmdLineInputError(
                "op settings must be defined on the base resource, "
                "not the group"
            )
        if len(group_groups.get_args_flat("group_id")) != 1:
            raise CmdLineInputError(
                "You have to specify exactly one group after 'group'"
            )
        position: dict[str, Optional[str]] = {"after": None, "before": None}
        for where in position:
            if group_groups.has_keyword(where):
                if len(group_groups.get_args_flat(where)) != 1:
                    raise CmdLineInputError(
                        f"You have to specify exactly one resource after '{where}'"
                    )
                position[where] = group_groups.get_args_flat(where)[0]
        group_options = GroupOptions(
            group_id=group_groups.get_args_flat("group_id")[0],
            before_resource=position["before"],
            after_resource=position["after"],
        )

    clone_options: dict[str, Optional[CloneOptions]] = {
        "clone": None,
        "promotable": None,
    }
    for clone_type in clone_options:
        if not top_groups.has_keyword(clone_type):
            continue
        clone_groups = group_by_keywords(
            top_groups.get_args_flat(clone_type),
            {"op", "meta"},
            implicit_first_keyword="options",
        )
        clone_id = None
        options = clone_groups.get_args_flat("options")
        if options and "=" not in options[0]:
            clone_id = options.pop(0)
        if options:
            raise CmdLineInputError(
                f"Specifying instance attributes for a {clone_type} "
                f"is not supported. Use 'meta' after '{clone_type}' "
                "if you want to specify meta attributes."
            )
        if clone_groups.has_keyword("op"):
            raise CmdLineInputError(
                "op settings must be defined on the base resource, "
                f"not the {clone_type}"
            )
        clone_options[clone_type] = CloneOptions(
            clone_id=clone_id,
            meta_attrs=KeyValueParser(
                clone_groups.get_args_flat("meta")
            ).get_unique(),
        )

    bundle_id = None
    if top_groups.has_keyword("bundle"):
        bundle_groups = group_by_keywords(
            top_groups.get_args_flat("bundle"),
            {"op", "meta"},
            implicit_first_keyword="options",
        )
        if bundle_groups.has_keyword("meta"):
            raise CmdLineInputError(
                "meta options must be defined on the base resource, "
                "not the bundle"
            )
        if bundle_groups.has_keyword("op"):
            raise CmdLineInputError(
                "op settings must be defined on the base resource, "
                "not the bundle"
            )
        if len(bundle_groups.get_args_flat("options")) != 1:
            raise CmdLineInputError("you have to specify exactly one bundle")
        bundle_id = bundle_groups.get_args_flat("options")[0]

    return ComplexResourceOptions(
        primitive=primitive_options,
        group=group_options,
        clone=clone_options["clone"],
        promotable=clone_options["promotable"],
        bundle_id=bundle_id,
    )


# deprecated since 0.11.6
def parse_create_old(  # noqa: PLR0912
    arg_list: Argv, modifiers: InputModifiers
) -> ComplexResourceOptions:
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    top_groups = group_by_keywords(
        arg_list,
        {"clone", "promotable", "bundle"},
        implicit_first_keyword="primitive",
    )

    primitive_groups = group_by_keywords(
        top_groups.get_args_flat("primitive"),
        {"op", "meta"},
        implicit_first_keyword="instance",
    )
    primitive_instance_attrs = primitive_groups.get_args_flat("instance")
    primitive_meta_attrs = primitive_groups.get_args_flat("meta")
    primitive_operations = primitive_groups.get_args_groups("op")

    group_options = None
    if modifiers.is_specified("--group"):
        dash_deprecation = (
            "Using '--{option}' is deprecated and will be replaced with "
            "'{option}' in a future release. "
            f"Specify {FUTURE_OPTION} to switch to the future behavior."
        )
        deprecation_warning(dash_deprecation.format(option="group"))
        before_resource = None
        if modifiers.get("--before"):
            before_resource = str(modifiers.get("--before"))
            deprecation_warning(dash_deprecation.format(option="before"))
        after_resource = None
        if modifiers.get("--after"):
            after_resource = str(modifiers.get("--after"))
            deprecation_warning(dash_deprecation.format(option="after"))
        group_options = GroupOptions(
            group_id=str(modifiers.get("--group")),
            before_resource=before_resource,
            after_resource=after_resource,
        )
    else:
        for option in ("--before", "--after"):
            if modifiers.is_specified(option):
                raise CmdLineInputError(
                    f"you cannot use {option} without --group"
                )

    clone_options: dict[str, Optional[CloneOptions]] = {
        "clone": None,
        "promotable": None,
    }
    for clone_type in clone_options:
        if not top_groups.has_keyword(clone_type):
            continue
        clone_groups = group_by_keywords(
            top_groups.get_args_flat(clone_type),
            {"op", "meta"},
            implicit_first_keyword="options",
        )
        clone_id = None
        options = clone_groups.get_args_flat("options")
        if options and "=" not in options[0]:
            clone_id = options.pop(0)
        if options:
            deprecation_warning(
                f"Configuring {clone_type} meta attributes without specifying "
                f"the 'meta' keyword after the '{clone_type}' keyword "
                "is deprecated and will be removed in a future release. "
                f"Specify {FUTURE_OPTION} to switch to the future behavior."
            )
        if clone_groups.has_keyword("op"):
            deprecation_warning(
                f"Specifying 'op' after '{clone_type}' now defines "
                "operations for the base resource. In a future release, this "
                f"will be removed and operations will have to be specified "
                f"before '{clone_type}'. "
                f"Specify {FUTURE_OPTION} to switch to the future behavior."
            )
            primitive_operations += clone_groups.get_args_groups("op")
        if clone_groups.has_keyword("meta"):
            deprecation_warning(
                f"Specifying 'meta' after '{clone_type}' now defines "
                "meta attributes for the base resource. In a future release, this "
                f"will define meta attributes for the {clone_type}. "
                f"Specify {FUTURE_OPTION} to switch to the future behavior."
            )
            primitive_meta_attrs += clone_groups.get_args_flat("meta")
        clone_options[clone_type] = CloneOptions(
            clone_id=clone_id,
            meta_attrs=KeyValueParser(options).get_unique(),
        )

    bundle_id = None
    if top_groups.has_keyword("bundle"):
        bundle_groups = group_by_keywords(
            top_groups.get_args_flat("bundle"),
            {"op", "meta"},
            implicit_first_keyword="options",
        )
        if bundle_groups.has_keyword("meta"):
            deprecation_warning(
                "Specifying 'meta' after 'bundle' now defines meta options for "
                "the base resource. In a future release, this will be removed and meta "
                "options will have to be specified before 'bundle'. "
                f"Specify {FUTURE_OPTION} to switch to the future behavior."
            )
            primitive_meta_attrs += bundle_groups.get_args_flat("meta")
        if bundle_groups.has_keyword("op"):
            deprecation_warning(
                "Specifying 'op' after 'bundle' now defines operations for the "
                "base resource. In a future release, this will be removed and operations "
                "will have to be specified before 'bundle'. "
                f"Specify {FUTURE_OPTION} to switch to the future behavior."
            )
            primitive_operations += bundle_groups.get_args_groups("op")
        if len(bundle_groups.get_args_flat("options")) != 1:
            raise CmdLineInputError("you have to specify exactly one bundle")
        bundle_id = bundle_groups.get_args_flat("options")[0]

    return ComplexResourceOptions(
        primitive=PrimitiveOptions(
            KeyValueParser(primitive_instance_attrs).get_unique(),
            KeyValueParser(primitive_meta_attrs).get_unique(),
            [
                KeyValueParser(op).get_unique()
                for op in build_operations(primitive_operations)
            ],
        ),
        group=group_options,
        clone=clone_options["clone"],
        promotable=clone_options["promotable"],
        bundle_id=bundle_id,
    )


def _parse_bundle_groups(arg_list: Argv) -> ArgsByKeywords:
    """
    Commandline options: no options
    """
    repeatable_keyword_list = ["port-map", "storage-map"]
    keyword_list = ["meta", "container", "network"] + repeatable_keyword_list
    groups = group_by_keywords(arg_list, set(keyword_list))
    for keyword in keyword_list:
        if not groups.has_keyword(keyword):
            continue
        if keyword in repeatable_keyword_list:
            for repeated_section in groups.get_args_groups(keyword):
                if not repeated_section:
                    raise CmdLineInputError(f"No {keyword} options specified")
        elif not groups.get_args_flat(keyword):
            raise CmdLineInputError(f"No {keyword} options specified")
    return groups


def _parse_bundle_create_or_reset(
    arg_list: Argv, reset: bool
) -> BundleCreateOptions:
    """
    Commandline options: no options
    """
    groups = _parse_bundle_groups(arg_list)
    container_options = groups.get_args_flat("container")
    container_type = ""
    if not reset and container_options and "=" not in container_options[0]:
        container_type = container_options.pop(0)
    return BundleCreateOptions(
        container_type=container_type,
        container=KeyValueParser(container_options).get_unique(),
        network=KeyValueParser(groups.get_args_flat("network")).get_unique(),
        port_map=[
            KeyValueParser(port_map).get_unique()
            for port_map in groups.get_args_groups("port-map")
        ],
        storage_map=[
            KeyValueParser(storage_map).get_unique()
            for storage_map in groups.get_args_groups("storage-map")
        ],
        meta_attrs=KeyValueParser(groups.get_args_flat("meta")).get_unique(),
    )


def parse_bundle_create_options(arg_list: Argv) -> BundleCreateOptions:
    """
    Commandline options: no options
    """
    return _parse_bundle_create_or_reset(arg_list, reset=False)


def parse_bundle_reset_options(arg_list: Argv) -> BundleCreateOptions:
    """
    Commandline options: no options
    """
    return _parse_bundle_create_or_reset(arg_list, reset=True)


def _split_bundle_map_update_op_and_options(
    map_arg_list: Argv, result_parts: AddRemoveOptions, map_name: str
) -> None:
    if len(map_arg_list) < 2:
        raise _bundle_map_update_not_valid(map_name)
    op, options = map_arg_list[0], map_arg_list[1:]
    if op == "add":
        result_parts.add.append(KeyValueParser(options).get_unique())
    elif op in {"delete", "remove"}:
        result_parts.remove.extend(options)
    else:
        raise _bundle_map_update_not_valid(map_name)


def _bundle_map_update_not_valid(map_name: str) -> CmdLineInputError:
    return CmdLineInputError(
        f"When using '{map_name}' you must specify either 'add' and options or "
        "either of 'delete' or 'remove' and id(s)"
    )


def parse_bundle_update_options(arg_list: Argv) -> BundleUpdateOptions:
    """
    Commandline options: no options
    """
    groups = _parse_bundle_groups(arg_list)
    port_map = AddRemoveOptions(add=[], remove=[])
    for map_group in groups.get_args_groups("port-map"):
        _split_bundle_map_update_op_and_options(map_group, port_map, "port-map")
    storage_map = AddRemoveOptions(add=[], remove=[])
    for map_group in groups.get_args_groups("storage-map"):
        _split_bundle_map_update_op_and_options(
            map_group, storage_map, "storage-map"
        )
    return BundleUpdateOptions(
        container=KeyValueParser(
            groups.get_args_flat("container")
        ).get_unique(),
        network=KeyValueParser(groups.get_args_flat("network")).get_unique(),
        port_map_add=port_map.add,
        port_map_remove=port_map.remove,
        storage_map_add=storage_map.add,
        storage_map_remove=storage_map.remove,
        meta_attrs=KeyValueParser(groups.get_args_flat("meta")).get_unique(),
    )


def build_operations(op_group_list: list[Argv]) -> list[Argv]:
    """
    Return a list of dicts. Each dict represents one operation.

    op_group_list -- contains items that have parameters after "op"
        (so item can contain multiple operations) for example: [
            [monitor timeout=1 start timeout=2],
            [monitor timeout=3 interval=10],
        ]
    """
    operation_list = []
    for op_group in op_group_list:
        # empty operation is not allowed
        if not op_group:
            raise __not_enough_parts_in_operation()

        # every operation group needs to start with operation name
        if "=" in op_group[0]:
            raise __every_operation_needs_name()

        for arg in op_group:
            if "=" not in arg:
                operation_list.append([f"name={arg}"])
            else:
                operation_list[-1].append(arg)

    # every operation needs at least name and one option
    # there can be more than one operation in op_group: check is after
    # processing
    if any(len(operation) < 2 for operation in operation_list):
        raise __not_enough_parts_in_operation()

    return operation_list


def __not_enough_parts_in_operation() -> CmdLineInputError:
    return CmdLineInputError(
        "When using 'op' you must specify an operation name"
        " and at least one option"
    )


def __every_operation_needs_name() -> CmdLineInputError:
    return CmdLineInputError(
        "When using 'op' you must specify an operation name after 'op'"
    )



from shlex import quote
from typing import (
    Optional,
    Sequence,
)

from pcs.cli.nvset import nvset_dto_to_lines
from pcs.cli.resource.output import resource_agent_parameter_metadata_to_text
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.nvset import (
    CibNvsetDto,
    ListCibNvsetDto,
)
from pcs.common.resource_agent.dto import ResourceAgentParameterDto
from pcs.common.str_tools import (
    format_name_value_default_list,
    format_name_value_list,
    indent,
)
from pcs.common.types import (
    StringCollection,
    StringSequence,
)


class PropertyConfigurationFacade:
    def __init__(
        self,
        properties: Sequence[CibNvsetDto],
        properties_metadata: Sequence[ResourceAgentParameterDto],
        readonly_properties: StringCollection,
    ) -> None:
        self._properties = properties
        self._first_nvpair_set = (
            self._properties[0].nvpairs if self._properties else []
        )
        self._properties_metadata = properties_metadata
        self._readonly_properties = readonly_properties
        self._defaults_map = self.get_defaults(include_advanced=True)
        self._name_nvpair_dto_map = {
            nvpair_dto.name: nvpair_dto for nvpair_dto in self._first_nvpair_set
        }

    @classmethod
    def from_properties_dtos(
        cls,
        properties_dto: ListCibNvsetDto,
        properties_metadata_dto: ClusterPropertyMetadataDto,
    ) -> "PropertyConfigurationFacade":
        return cls(
            properties_dto.nvsets,
            properties_metadata_dto.properties_metadata,
            properties_metadata_dto.readonly_properties,
        )

    @classmethod
    def from_properties_config(
        cls, properties_dto: ListCibNvsetDto
    ) -> "PropertyConfigurationFacade":
        return cls(
            properties_dto.nvsets,
            [],
            [],
        )

    @classmethod
    def from_properties_metadata(
        cls, properties_metadata_dto: ClusterPropertyMetadataDto
    ) -> "PropertyConfigurationFacade":
        return cls(
            [],
            properties_metadata_dto.properties_metadata,
            properties_metadata_dto.readonly_properties,
        )

    @property
    def properties(self) -> Sequence[CibNvsetDto]:
        return self._properties

    @property
    def properties_metadata(self) -> Sequence[ResourceAgentParameterDto]:
        return self._properties_metadata

    @property
    def readonly_properties(self) -> StringCollection:
        return self._readonly_properties

    def get_property_value(
        self, property_name: str, custom_default: Optional[str] = None
    ) -> Optional[str]:
        nvpair = self._name_nvpair_dto_map.get(property_name)
        return nvpair.value if nvpair else custom_default

    def get_property_value_or_default(
        self, property_name: str, custom_default: Optional[str] = None
    ) -> Optional[str]:
        value = self.get_property_value(property_name)
        if value is not None:
            return value
        return self._defaults_map.get(property_name, custom_default)

    def get_defaults(
        self,
        property_names: Optional[StringSequence] = None,
        include_advanced: bool = False,
    ) -> dict[str, str]:
        return {
            metadata.name: metadata.default
            for metadata in self.get_properties_metadata(
                property_names, include_advanced
            )
            if metadata.default is not None
        }

    def get_properties_metadata(
        self,
        property_names: Optional[StringSequence] = None,
        include_advanced: bool = False,
    ) -> Sequence[ResourceAgentParameterDto]:
        if property_names:
            filtered_metadata = [
                metadata
                for metadata in self._properties_metadata
                if metadata.name in property_names
            ]
        else:
            filtered_metadata = [
                metadata
                for metadata in self._properties_metadata
                if include_advanced or not metadata.advanced
            ]
        deduplicated_metadata = {
            metadata.name: metadata for metadata in filtered_metadata
        }
        return list(deduplicated_metadata.values())

    def get_name_value_default_list(self) -> list[tuple[str, str, bool]]:
        name_value_default_list = [
            (nvpair_dto.name, nvpair_dto.value, False)
            for nvpair_dto in self._first_nvpair_set
        ]
        name_value_default_list.extend(
            [
                (metadata_dto.name, metadata_dto.default, True)
                for metadata_dto in self.get_properties_metadata(
                    include_advanced=True
                )
                if metadata_dto.name not in self._name_nvpair_dto_map
                and metadata_dto.default is not None
            ]
        )
        return name_value_default_list


def properties_to_text(
    properties_facade: PropertyConfigurationFacade,
) -> list[str]:
    """
    Return a text format of configured properties.

    properties_facade -- cluster property configuration and metadata
    """
    if properties_facade.properties:
        return nvset_dto_to_lines(
            properties_facade.properties[0],
            nvset_label="Cluster Properties",
        )
    return []


def properties_to_text_with_default_mark(
    properties_facade: PropertyConfigurationFacade,
    property_names: Optional[StringSequence] = None,
) -> list[str]:
    """
    Return text format of configured properties or property default values.
    If default property value is missing then property is not displayed at all.
    If property_names is specified, then only properties from the list is
    displayed.

    properties_facade -- cluster property configuration and metadata
    property_names -- properties to be displayed
    """
    lines: list[str] = []
    id_part = (
        f" {properties_facade.properties[0].id}"
        if properties_facade.properties
        else ""
    )
    lines = [f"Cluster Properties:{id_part}"]
    tuple_list = [
        item
        for item in properties_facade.get_name_value_default_list()
        if not property_names or item[0] in property_names
    ]
    lines.extend(indent(format_name_value_default_list(sorted(tuple_list))))
    return lines


def properties_to_cmd(
    properties_facade: PropertyConfigurationFacade,
) -> list[str]:
    """
    Convert configured properties to the `pcs property set` command.

    properties_facade -- cluster property configuration and metadata
    """
    if properties_facade.properties and properties_facade.properties[0].nvpairs:
        options = [
            quote("=".join([nvpair.name, nvpair.value]))
            for nvpair in properties_facade.properties[0].nvpairs
            if nvpair.name not in properties_facade.readonly_properties
        ]
        if options:
            return ["pcs property set --force --"] + indent(options)
    return []


def properties_defaults_to_text(property_dict: dict[str, str]) -> list[str]:
    """
    Convert property default values to lines of text.

    property_dict -- name to default value map
    """
    return format_name_value_list(sorted(property_dict.items()))


def cluster_property_metadata_to_text(
    metadata: Sequence[ResourceAgentParameterDto],
) -> list[str]:
    """
    Convert cluster property metadata to lines of description text.
    Output example:

    property-name
      Description: <longdesc or shortdesc>
      Type: <type> / Allowed values: <enum values>
      Default: <default value>

    metadata - list of ResourceAgentParameterDto which is used for cluster
        property metadata
    """
    text: list[str] = []
    for parameter_dto in metadata:
        text.extend(resource_agent_parameter_metadata_to_text(parameter_dto))
    return text



from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import CibRuleExpressionType

from .expression_part import BoolExpr, RscExpr, RuleExprPart


def is_rsc_expressions_only(rule_expr: RuleExprPart) -> bool:
    if isinstance(rule_expr, RscExpr):
        return True
    if isinstance(rule_expr, BoolExpr):
        return all(
            is_rsc_expressions_only(child) for child in rule_expr.children
        )
    return False


def is_rsc_expressions_only_dto(rule_dto: CibRuleExpressionDto) -> bool:
    if rule_dto.type == CibRuleExpressionType.RSC_EXPRESSION:
        return True
    if rule_dto.type == CibRuleExpressionType.RULE:
        return all(
            is_rsc_expressions_only_dto(child) for child in rule_dto.expressions
        )
    return False



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.resource import command
from pcs.common import reports

from pcs_test.tools.misc import dict_to_modifiers

RESOURCE_ID = "resource-id"


class ResourceMeta(TestCase):
    command = staticmethod(command.meta)

    def setUp(self):
        self.lib = mock.Mock(spec_set=["resource", "cluster"])

        self.lib.resource = mock.Mock(
            spec_set=[
                "update_meta",
                "is_any_stonith",
                "is_any_resource_except_stonith",
            ]
        )
        self.lib.resource.update_meta = self.update_meta = mock.Mock()

        mock_get_resource_status_msg_patcher = mock.patch(
            "pcs.cli.resource.command.get_resource_status_msg"
        )
        self.addCleanup(mock_get_resource_status_msg_patcher.stop)
        self.mock_get_resource_status_msg = (
            mock_get_resource_status_msg_patcher.start()
        )

        self.lib.resource.is_any_stonith = mock.Mock()
        self.lib.resource.is_any_stonith.return_value = False

        self.lib.cluster = mock.Mock(spec_set=["wait_for_pcmk_idle"])
        self.lib.cluster.wait_for_pcmk_idle = self.wait_for_pcmk_idle = (
            mock.Mock()
        )

        mock_deprecation_warning_patcher = mock.patch(
            "pcs.cli.resource.command.deprecation_warning"
        )
        self.addCleanup(mock_deprecation_warning_patcher.stop)
        self.mock_deprecation_warning = mock_deprecation_warning_patcher.start()

    def _assert_no_wait_or_stonith(self):
        self.lib.resource.is_any_stonith.assert_called_once_with([RESOURCE_ID])
        self.wait_for_pcmk_idle.assert_not_called()
        self.mock_get_resource_status_msg.assert_not_called()
        self.mock_deprecation_warning.assert_not_called()

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.command(self.lib, [], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.resource.is_any_stonith.assert_not_called()
        self.update_meta.assert_not_called()
        self.wait_for_pcmk_idle.assert_not_called()

    def test_duplicate_attrs(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.command(
                self.lib,
                [RESOURCE_ID, "meta-attr1=value1", "meta-attr1=value2"],
                dict_to_modifiers({}),
            )
        self.assertEqual(
            cm.exception.message,
            "duplicate option 'meta-attr1' with different values 'value1' and "
            "'value2'",
        )
        self.update_meta.assert_not_called()
        self._assert_no_wait_or_stonith()

    def test_one_attr(self):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1"],
            dict_to_modifiers({}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1"}, []
        )
        self._assert_no_wait_or_stonith()

    @mock.patch("pcs.cli.reports.output.print_to_stderr")
    def test_stonith_forbidden(self, mock_print):
        self.lib.resource.is_any_stonith.return_value = True
        self.assertRaises(
            SystemExit,
            lambda: self.command(
                self.lib,
                [RESOURCE_ID, "meta-attr1=value1"],
                dict_to_modifiers({}),
            ),
        )
        self.update_meta.assert_not_called()
        mock_print.assert_called_once_with(
            "Error: This command does not accept stonith resources. Please use "
            "'pcs stonith meta' instead."
        )
        self._assert_no_wait_or_stonith()

    def test_multiple_attrs(self):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1", "meta-attr2=value2"],
            dict_to_modifiers({}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1", "meta-attr2": "value2"}, []
        )
        self._assert_no_wait_or_stonith()

    @mock.patch("pcs.cli.resource.command.print")
    def test_with_wait_zero(self, _):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1"],
            dict_to_modifiers({"wait": "0"}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1"}, []
        )
        self.lib.resource.is_any_stonith.assert_called_once_with([RESOURCE_ID])
        self.wait_for_pcmk_idle.assert_called_once_with(0)
        self.mock_deprecation_warning.assert_called_once_with(
            "Ability of this command to accept 'wait' argument is "
            "deprecated and will be removed in a future release."
        )

    @mock.patch("pcs.cli.resource.command.print")
    @mock.patch("pcs.cli.stonith.command.print")
    def test_with_wait_timeout(self, _a, _b):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1"],
            dict_to_modifiers({"wait": "30"}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1"}, []
        )
        self.wait_for_pcmk_idle.assert_called_once_with(30)
        self.mock_get_resource_status_msg.assert_called_once_with(
            mock.ANY, RESOURCE_ID
        )
        self.mock_deprecation_warning.assert_called_once_with(
            "Ability of this command to accept 'wait' argument is "
            "deprecated and will be removed in a future release."
        )

    def test_with_force(self):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1"],
            dict_to_modifiers({"force": True}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1"}, [reports.codes.FORCE]
        )
        self.mock_get_resource_status_msg.assert_not_called()
        self._assert_no_wait_or_stonith()



import logging
from unittest import mock

from tornado.httputil import parse_cookie

from pcs.daemon import ruby_pcsd
from pcs.daemon.app import sinatra_ui
from pcs.daemon.app.auth import UnixSocketAuthProvider

try:
    from pcs.daemon.app import webui
except ImportError:
    webui = None
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from pcs_test.tier0.daemon.app import (
    fixtures_app,
    fixtures_app_webui,
)
from pcs_test.tools.misc import skip_unless_webui_installed

# Don't write errors to test output.
logging.getLogger("tornado.access").setLevel(logging.CRITICAL)


def patch_login_user():
    return mock.patch.object(
        AuthProvider,
        "login_user",
        lambda _self, username: AuthUser(username=username, groups=[username]),
    )


def patch_get_unix_socket_user(user):
    return mock.patch.object(
        UnixSocketAuthProvider,
        "get_unix_socket_user",
        lambda _self: user,
    )


auth_provider = AuthProvider(logging.getLogger("test logger"))


@skip_unless_webui_installed()
@patch_login_user()
class SinatraAjaxProtectedSession(fixtures_app_webui.AppTest):
    def setUp(self):
        self.wrapper = fixtures_app.RubyPcsdWrapper(ruby_pcsd.SINATRA)
        super().setUp()

    def get_routes(self):
        return webui.sinatra_ui.get_routes(
            self.session_storage,
            auth_provider,
            self.wrapper,
        )

    def assert_session_in_response(self, response, sid=None):
        self.assertTrue("Set-Cookie" in response.headers)
        cookie = parse_cookie(response.headers["Set-Cookie"])
        self.assertTrue(webui.auth.PCSD_SESSION, cookie)
        if sid:
            self.assertEqual(cookie[webui.auth.PCSD_SESSION], sid)
        return cookie[webui.auth.PCSD_SESSION]

    def test_deal_without_authentication(self):
        self.assert_unauth_ajax(self.get("/some-ajax", is_ajax=True))

    def test_not_ajax(self):
        session1 = self.create_login_session()
        self.assert_unauth_ajax(self.get("/some-ajax", sid=session1.sid))

    def test_take_result_from_ruby(self):
        session1 = self.create_login_session()
        response = self.get("/some-ajax", sid=session1.sid, is_ajax=True)
        self.assert_wrappers_response(response)
        self.assert_session_in_response(response, session1.sid)


@patch_login_user()
class SinatraAjaxProtectedUnixSocket(fixtures_app.AppTest):
    def setUp(self):
        self.wrapper = fixtures_app.RubyPcsdWrapper(ruby_pcsd.SINATRA)
        super().setUp()

    def get_routes(self):
        return sinatra_ui.get_routes(auth_provider, self.wrapper)

    @patch_get_unix_socket_user(None)
    def test_deal_without_authentication(self):
        self.assert_unauth_ajax(self.get("/some-ajax", is_ajax=True))

    @patch_get_unix_socket_user(fixtures_app.USER)
    def test_take_result_from_ruby(self):
        response = self.get("/some-ajax", is_ajax=True)
        self.assert_wrappers_response(response)

    @patch_get_unix_socket_user(fixtures_app.USER)
    def test_not_ajax(self):
        self.assert_unauth_ajax(self.get("/some-ajax"))



from dataclasses import dataclass
from typing import Optional, Sequence

from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto


@dataclass(frozen=True)
class CibNodeDto(DataTransferObject):
    id: str
    uname: str
    description: Optional[str]
    score: Optional[str]
    type: Optional[str]
    instance_attributes: Sequence[CibNvsetDto]
    utilization: Sequence[CibNvsetDto]


@dataclass(frozen=True)
class CibNodeListDto(DataTransferObject):
    nodes: Sequence[CibNodeDto]



from typing import Union

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    KeyValueParser,
    split_list,
)


def prepare_resource_sets(
    cmdline_args: Argv,
) -> list[dict[str, Union[list[str], dict[str, str]]]]:
    return [
        {
            "ids": [id_ for id_ in args if "=" not in id_],
            "options": KeyValueParser(
                [opt for opt in args if "=" in opt]
            ).get_unique(),
        }
        for args in split_list(cmdline_args, "set")
    ]


def prepare_set_args(
    argv: Argv,
) -> tuple[list[dict[str, Union[list[str], dict[str, str]]]], dict[str, str]]:
    args_groups = split_list(argv, "setoptions")
    if len(args_groups) > 2:
        raise CmdLineInputError(
            "Keyword 'setoptions' may be mentioned at most once"
        )
    resource_set_args = args_groups[0]
    constraint_options_args = args_groups[1] if len(args_groups) == 2 else []

    if not resource_set_args:
        raise CmdLineInputError()

    resource_set_list = prepare_resource_sets(resource_set_args)
    if not resource_set_list or not all(
        resource_set["ids"] for resource_set in resource_set_list
    ):
        raise CmdLineInputError()

    constraint_options = {}
    if constraint_options_args:
        constraint_options = KeyValueParser(
            constraint_options_args
        ).get_unique()

    return (resource_set_list, constraint_options)



from dataclasses import dataclass
from typing import Sequence

from pcs.common.interface.dto import DataTransferObject
from pcs.common.resource_agent.dto import ResourceAgentParameterDto
from pcs.common.types import StringCollection


@dataclass(frozen=True)
class ClusterPropertyMetadataDto(DataTransferObject):
    properties_metadata: Sequence[ResourceAgentParameterDto]
    readonly_properties: StringCollection



from typing import Mapping, Optional, cast

from lxml.etree import SubElement, _Element

from pcs.common import const, reports
from pcs.common.pacemaker.constraint import (
    CibConstraintTicketAttributesDto,
    CibConstraintTicketDto,
    CibConstraintTicketSetDto,
)
from pcs.common.pacemaker.role import (
    get_value_for_cib as get_role_value_for_cib,
)
from pcs.common.pacemaker.types import CibTicketLossPolicy
from pcs.lib import validate
from pcs.lib.booth.config_validators import validate_ticket_name
from pcs.lib.cib.const import TAG_CONSTRAINT_TICKET as TAG
from pcs.lib.cib.tools import IdProvider, Version, role_constructor
from pcs.lib.pacemaker.values import sanitize_id
from pcs.lib.tools import get_optional_value
from pcs.lib.xml_tools import remove_one_element, remove_when_pointless

from .common import (
    CmdInputResourceSetList,
    CmdInputResourceSetLoadedList,
    DuplicatesChecker,
    DuplicatesCheckerSetConstraint,
    create_constraint_with_set,
    is_set_constraint,
    validate_constrainable_elements,
)
from .resource_set import constraint_element_to_resource_set_dto_list

_LOSS_POLICY_VALUES = ("fence", "stop", "freeze", "demote")


def is_ticket_constraint(element: _Element) -> bool:
    return element.tag == TAG


def _validate_ticket(ticket: Optional[str]) -> reports.ReportItemList:
    # use a booth ticket validator for validating a ticket
    if not ticket:
        return [
            reports.ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["ticket"])
            )
        ]
    return validate_ticket_name(ticket)


def validate_create_plain(
    id_provider: IdProvider,
    ticket: str,
    constrained_el: Optional[_Element],
    options: validate.TypeOptionMap,
    in_multiinstance_allowed: bool,
) -> reports.ReportItemList:
    """
    Validator for creating new plain ticket constraint

    id_provider -- elements' ids generator
    ticket -- name of the ticket
    constrained_el -- an element to be constrained
    options -- additional options for the constraint
    in_multiinstance_allowed -- allow constraints for resources in clones/bundles
    """
    report_list: reports.ReportItemList = []

    # validate resource specification
    # caller is responsible for handling the 'resource not found' case
    if constrained_el is not None:
        report_list.extend(
            validate_constrainable_elements(
                [constrained_el], in_multiinstance_allowed
            )
        )

    # validate options
    report_list += _validate_ticket(ticket)
    validators = [
        validate.NamesIn(
            # rsc and ticket are passed as parameters, not as items in the
            # options dict
            {"id", "loss-policy", "rsc-role"}
        ),
        # with id_provider it validates that the id is available as well
        validate.ValueId(
            "id",
            option_name_for_report="constraint id",
            id_provider=id_provider,
        ),
        validate.ValueIn(
            "rsc-role", const.PCMK_ROLES, option_name_for_report="role"
        ),
        validate.ValueIn("loss-policy", _LOSS_POLICY_VALUES),
    ]
    report_list.extend(validate.ValidatorAll(validators).validate(options))

    return report_list


def create_plain(
    parent_element: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    ticket: str,
    resource_id: str,
    options: Mapping[str, str],
) -> _Element:
    """
    Create a plain ticket constraint

    parent_element -- where to place the constraint
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    ticket -- name of the ticket
    resource_id -- resource to be constrained
    options -- additional options for the constraint
    """
    options = dict(options)  # make a modifiable copy

    # prepare resource role
    if "rsc-role" in options:
        options["rsc-role"] = get_role_value_for_cib(
            const.PcmkRoleType(options["rsc-role"]),
            cib_schema_version >= const.PCMK_NEW_ROLES_CIB_VERSION,
        )
    resource_role = options.get("rsc-role", "")

    # autogenerate id if not provided
    if "id" not in options:
        options["id"] = id_provider.allocate_id(
            sanitize_id(
                "-".join(["ticket", ticket, resource_id])
                + (f"-{resource_role}" if resource_role else ""),
            )
        )

    # create the constraint element
    options["ticket"] = ticket
    options["rsc"] = resource_id
    constraint_el = SubElement(parent_element, TAG)
    for name, value in options.items():
        if value != "":
            constraint_el.attrib[name] = value

    return constraint_el


def validate_create_with_set(
    id_provider: IdProvider,
    rsc_set_list: CmdInputResourceSetLoadedList,
    options: validate.TypeOptionMap,
    in_multiinstance_allowed: bool,
) -> reports.ReportItemList:
    """
    Validator for creating new ticket constraint with resource sets

    id_provider -- elements' ids generator
    rsc_set_list -- definition of resource sets: resources and set options
    options -- additional options for the constraint
    in_multiinstance_allowed -- allow constraints for resources in clones/bundles
    """
    report_list: reports.ReportItemList = []

    # TODO Set options validators were moved here during refactoring. No
    # changes were done to the validators, as that would change pcs behavior,
    # which we were avoiding during refactoring. It is, however, desired to put
    # correct validators in place.
    set_options_validators = [
        validate.NamesIn(
            ("action", "require-all", "role", "sequential"), option_type="set"
        ),
        validate.ValueIn("action", const.PCMK_ACTIONS),
        validate.ValuePcmkBoolean("require-all"),
        validate.ValueIn("role", const.PCMK_ROLES),
        validate.ValuePcmkBoolean("sequential"),
    ]
    # validate resources specification and resource set options
    # caller is responsible for handling the 'resource not found' case
    if not rsc_set_list:
        report_list.append(
            reports.ReportItem.error(reports.messages.EmptyResourceSetList())
        )
    for rsc_set in rsc_set_list:
        if rsc_set["constrained_elements"]:
            report_list.extend(
                validate_constrainable_elements(
                    rsc_set["constrained_elements"], in_multiinstance_allowed
                )
            )
        else:
            report_list.append(
                reports.ReportItem.error(reports.messages.EmptyResourceSet())
            )
        report_list.extend(
            validate.ValidatorAll(set_options_validators).validate(
                rsc_set["options"]
            )
        )

    # validate options
    ticket = validate.pairs_to_values(options).get("ticket")
    report_list += _validate_ticket(ticket)
    validators = [
        validate.NamesIn({"id", "loss-policy", "ticket"}),
        # with id_provider it validates that the id is available as well
        validate.ValueId(
            "id",
            option_name_for_report="constraint id",
            id_provider=id_provider,
        ),
        validate.ValueIn("loss-policy", _LOSS_POLICY_VALUES),
    ]
    report_list.extend(validate.ValidatorAll(validators).validate(options))

    return report_list


def create_with_set(
    parent_element: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    rsc_set_list: CmdInputResourceSetList,
    options: Mapping[str, str],
) -> _Element:
    """
    Create a ticket constraint with resource sets

    parent_element -- where to place the constraint
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    rsc_set_list -- definition of resource sets: resources and set options
    options -- additional options for the constraint
    """
    return create_constraint_with_set(
        parent_element,
        id_provider,
        cib_schema_version,
        TAG,
        "ticket",
        rsc_set_list,
        options,
    )


def remove_plain(
    constraint_section: _Element, ticket_key: str, resource_id: str
) -> bool:
    ticket_element_list = cast(
        list[_Element],
        constraint_section.xpath(
            ".//rsc_ticket[@ticket=$ticket and @rsc=$resource]",
            ticket=ticket_key,
            resource=resource_id,
        ),
    )

    for ticket_element in ticket_element_list:
        remove_one_element(ticket_element)

    return len(ticket_element_list) > 0


def remove_with_resource_set(
    constraint_section: _Element, ticket_key: str, resource_id: str
) -> bool:
    ref_element_list = cast(
        list[_Element],
        constraint_section.xpath(
            ".//rsc_ticket[@ticket=$ticket]/resource_set/resource_ref[@id=$resource]",
            ticket=ticket_key,
            resource=resource_id,
        ),
    )

    for ref_element in ref_element_list:
        set_element = ref_element.getparent()
        if set_element is None:
            continue
        set_element.remove(ref_element)
        # set_element is lxml element, therefore we have to use len() here
        # pylint: disable=len-as-condition
        if not len(set_element):
            ticket_element = set_element.getparent()
            if ticket_element is None:
                continue
            ticket_element.remove(set_element)
            # We do not care about attributes since without an attribute "rsc"
            # they are pointless. Attribute "rsc" is mutually exclusive with
            # resource_set (see rng) so it cannot be in this ticket_element.
            remove_when_pointless(ticket_element, attribs_important=False)

    return len(ref_element_list) > 0


class DuplicatesCheckerTicketPlain(DuplicatesChecker):
    """
    Searcher of duplicate plain ticket constraints
    """

    _constraint_characteristics: Mapping[str, Optional[str]]

    @staticmethod
    def _characteristics(constraint_el: _Element) -> dict[str, Optional[str]]:
        return {
            "ticket": constraint_el.get("ticket"),
            "rsc": constraint_el.get("rsc"),
            "rsc-role": get_optional_value(
                role_constructor, constraint_el.get("rsc-role")
            ),
        }

    def _check_init(self, constraint_to_check: _Element) -> None:
        self._constraint_characteristics = self._characteristics(
            constraint_to_check
        )

    def _are_duplicate(
        self, constraint_to_check: _Element, constraint_el: _Element
    ) -> bool:
        del constraint_to_check
        return (
            self._characteristics(constraint_el)
            == self._constraint_characteristics
        )


class DuplicatesCheckerTicketWithSet(DuplicatesCheckerSetConstraint):
    """
    Searcher of duplicate ticket constraints with resource sets
    """

    def _are_duplicate(
        self,
        constraint_to_check: _Element,
        constraint_el: _Element,
    ) -> bool:
        return constraint_to_check.get("ticket") == constraint_el.get(
            "ticket"
        ) and super()._are_duplicate(constraint_to_check, constraint_el)


def _element_to_attributes_dto(
    element: _Element,
) -> CibConstraintTicketAttributesDto:
    return CibConstraintTicketAttributesDto(
        constraint_id=str(element.attrib["id"]),
        ticket=str(element.attrib["ticket"]),
        loss_policy=get_optional_value(
            CibTicketLossPolicy, element.get("loss-policy")
        ),
    )


def _constraint_el_to_dto(element: _Element) -> CibConstraintTicketDto:
    return CibConstraintTicketDto(
        resource_id=str(element.attrib["rsc"]),
        role=get_optional_value(role_constructor, element.get("rsc-role")),
        attributes=_element_to_attributes_dto(element),
    )


def _set_constraint_el_to_dto(element: _Element) -> CibConstraintTicketSetDto:
    return CibConstraintTicketSetDto(
        resource_sets=constraint_element_to_resource_set_dto_list(element),
        attributes=_element_to_attributes_dto(element),
    )


def get_all_as_dtos(
    constraints_el: _Element,
) -> tuple[list[CibConstraintTicketDto], list[CibConstraintTicketSetDto]]:
    plain_list: list[CibConstraintTicketDto] = []
    set_list: list[CibConstraintTicketSetDto] = []
    for constraint_el in constraints_el.findall(f"./{TAG}"):
        if is_set_constraint(constraint_el):
            set_list.append(_set_constraint_el_to_dto(constraint_el))
        else:
            plain_list.append(_constraint_el_to_dto(constraint_el))
    return plain_list, set_list



from unittest import TestCase, mock

from lxml import etree

from pcs.common import reports
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.lib import cluster_property as lib_cluster_property
from pcs.lib.cib.tools import IdProvider
from pcs.lib.pacemaker import values as packemaker_values
from pcs.lib.resource_agent import (
    ResourceAgentFacade,
    ResourceAgentMetadata,
    ResourceAgentName,
    ResourceAgentParameter,
)
from pcs.lib.resource_agent import const as ra_const
from pcs.lib.xml_tools import etree_to_str

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import get_runner_mock

TEMPLATE_CRM_CONFIG = """
<cib validate-with="pacemaker-3.8">
    <configuration>
        <crm_config>
            {property_sets}
        </crm_config>
    </configuration>
</cib>
"""

FIXTURE_TWO_PROPERTY_SETS = TEMPLATE_CRM_CONFIG.format(
    property_sets="""
        <cluster_property_set id="first">
            <nvpair id="first-have-watchdog" name="have-watchdog"
                value="false"/>
        </cluster_property_set>
        <cluster_property_set id="second">
            <nvpair id="second-maintenance-mode" name="maintenance-mode"
                value="false"/>
        </cluster_property_set>
    """
)


FIXTURE_NO_SET = TEMPLATE_CRM_CONFIG.format(property_sets="")

FORBIDDEN_OPTIONS_LIST = [
    "cluster-infrastructure",
    "cluster-name",
    "dc-version",
    "have-watchdog",
    "last-lrm-refresh",
]

PARAMETER_DEFINITIONS = [
    ("bool_param", "bool", "false", None, []),
    ("integer_param", "integer", "9", None, []),
    ("percentage_param", "percentage", "80%", None, []),
    ("select_param", "select", "s1", ["s1", "s2", "s3"], []),
    ("time_param", "time", "30s", None, []),
    (
        "stonith-watchdog-timeout",
        "time",
        "0",
        None,
        ["fencing-watchdog-timeout"],
    ),
    ("fencing-watchdog-timeout", "time", "0", None, []),
    ("stonith-enabled", "boolean", "true", None, []),
    ("fencing-enabled", "boolean", "true", None, []),
    ("cluster-infrastructure", "string", "corosync", None, []),
    ("cluster-name", "string", "(null)", None, []),
    ("dc-version", "string", "none", None, []),
    ("have-watchdog", "boolean", "false", None, []),
    ("duration_param", "duration", "P15S", None, []),
    ("nonnegative_param", "nonnegative_integer", "9", None, []),
    ("port_param", "port", "1234", None, []),
    ("score_param", "score", "INFINITY", None, []),
    ("timeout_param", "timeout", "60s", None, []),
]

ALLOWED_PROPERTIES = [
    "bool_param",
    "duration_param",
    "fencing-enabled",
    "fencing-watchdog-timeout",
    "integer_param",
    "nonnegative_param",
    "percentage_param",
    "port_param",
    "score_param",
    "select_param",
    "stonith-enabled",
    "stonith-watchdog-timeout",
    "time_param",
    "timeout_param",
]

FIXTURE_VALID_OPTIONS_DICT = {
    "bool_param": "true",
    "duration_param": "P1S",
    "integer_param": "10",
    "nonnegative_param": "9",
    "percentage_param": "20%",
    "port_param": "1234",
    "score_param": "-INFINITY",
    "select_param": "s3",
    "time_param": "5min",
    "timeout_param": "5",
}

FIXTURE_INVALID_OPTIONS_DICT = {
    "bool_param": "Falsch",
    "duration_param": "Pxy",
    "integer_param": "3.14",
    "nonnegative_param": "-1",
    "percentage_param": "20",
    "port_param": "70000",
    "score_param": "xyz",
    "select_param": "not-in-enum-values",
    "time_param": "10x",
    "timeout_param": "P1S",
    "unknown": "value",
    "have-watchdog": "100",
}

STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES = ["", "0", "0s"]
STONITH_WATCHDOG_TIMEOUT_PROPERTIES = [
    "stonith-watchdog-timeout",
    "fencing-watchdog-timeout",
]


def _fixture_parameter(name, param_type, default, enum_values, deprecated_by):
    return ResourceAgentParameter(
        name,
        shortdesc=None,
        longdesc=None,
        type=param_type,
        default=default,
        enum_values=enum_values,
        required=False,
        advanced=False,
        deprecated=bool(deprecated_by),
        deprecated_by=deprecated_by,
        deprecated_desc=None,
        unique_group=None,
        reloadable=False,
    )


FIXTURE_PARAMETER_LIST = [
    _fixture_parameter(*parameter) for parameter in PARAMETER_DEFINITIONS
]


FIXTURE_WATCHDOG_TIMEOUT_DEPRECATION_REPORT = fixture.warn(
    reports.codes.DEPRECATED_OPTION,
    option_name="stonith-watchdog-timeout",
    replaced_by=["fencing-watchdog-timeout"],
    option_type="cluster property",
)
FIXTURE_ERROR_REPORTS = [
    fixture.error(
        reports.codes.INVALID_OPTIONS,
        force_code=reports.codes.FORCE,
        option_names=["unknown"],
        allowed=ALLOWED_PROPERTIES,
        option_type="cluster property",
        allowed_patterns=[],
    ),
    fixture.error(
        reports.codes.INVALID_OPTIONS,
        force_code=None,
        option_names=["have-watchdog"],
        allowed=ALLOWED_PROPERTIES,
        option_type="cluster property",
        allowed_patterns=[],
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="duration_param",
        option_value="Pxy",
        allowed_values="time interval (e.g. 1, 2s, 3m, 4h, PT1H2M3S, ...)",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="integer_param",
        option_value="3.14",
        allowed_values="an integer or INFINITY or -INFINITY",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="nonnegative_param",
        option_value="-1",
        allowed_values="a positive integer or INFINITY",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="percentage_param",
        option_value="20",
        allowed_values=(
            "a non-negative integer followed by '%' (e.g. 0%, 50%, 200%, ...)"
        ),
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="port_param",
        option_value="70000",
        allowed_values=("a port number (1..65535)"),
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="score_param",
        option_value="xyz",
        allowed_values="an integer or INFINITY or -INFINITY",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="select_param",
        option_value="not-in-enum-values",
        allowed_values=["s1", "s2", "s3"],
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="time_param",
        option_value="10x",
        allowed_values="time interval (e.g. 1, 2s, 3m, 4h, PT1H2M3S, ...)",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="timeout_param",
        option_value="P1S",
        allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
]


def warning_reports(report_list):
    warning_report_list = []
    for item_tuple in report_list:
        item_list = list(item_tuple)
        if item_list[3] is not None:
            item_list[0] = reports.ReportItemSeverity.WARNING
            item_list[3] = None
        warning_report_list.append(tuple(item_list))
    return warning_report_list


class TestValidateSetClusterProperties(TestCase):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.mock_service_manager = mock.Mock(spec=ServiceManagerInterface)
        self.facade = ResourceAgentFacade(
            ResourceAgentMetadata(
                ResourceAgentName(
                    standard=ra_const.FAKE_AGENT_STANDARD,
                    provider=None,
                    type=ra_const.CLUSTER_OPTIONS,
                ),
                True,
                ra_const.OCF_1_1,
                "shortdesc",
                "longdesc",
                FIXTURE_PARAMETER_LIST,
                [],
            )
        )

        self.patcher_is_sbd_enabled = mock.patch("pcs.lib.sbd.is_sbd_enabled")
        self.patcher_sbd_devices = mock.patch(
            "pcs.lib.sbd.get_local_sbd_device_list"
        )
        self.patcher_sbd_timeout = mock.patch(
            "pcs.lib.sbd._get_local_sbd_watchdog_timeout"
        )
        self.mock_is_sbd_enabled = self.patcher_is_sbd_enabled.start()
        self.mock_sbd_devices = self.patcher_sbd_devices.start()
        self.mock_sbd_timeout = self.patcher_sbd_timeout.start()

    def tearDown(self):
        self.patcher_is_sbd_enabled.stop()
        self.patcher_sbd_devices.stop()
        self.patcher_sbd_timeout.stop()

    def assert_validate_set(
        self,
        configured_properties,
        new_properties,
        expected_report_list,
        *,
        sbd_enabled=False,
        sbd_devices=False,
        force=False,
        valid_value=True,
        returncode=0,
    ):
        calls_is_sbd_enabled = []
        calls_sbd_devices = []
        calls_sbd_timeout = []
        self.mock_is_sbd_enabled.reset_mock()
        self.mock_sbd_devices.reset_mock()
        self.mock_sbd_timeout.reset_mock()

        self.mock_is_sbd_enabled.return_value = sbd_enabled
        self.mock_sbd_devices.return_value = ["devices"] if sbd_devices else []
        self.mock_sbd_timeout.return_value = 10
        if "stonith-watchdog-timeout" in new_properties:
            expected_report_list = [
                FIXTURE_WATCHDOG_TIMEOUT_DEPRECATION_REPORT
            ] + expected_report_list

        assert_report_item_list_equal(
            lib_cluster_property.validate_set_cluster_properties(
                get_runner_mock(returncode=returncode),
                self.facade,
                "property-set-id",
                configured_properties,
                new_properties,
                self.mock_service_manager,
                force=force,
            ),
            expected_report_list,
        )

        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            if (
                prop_name in new_properties
                and (
                    new_properties[prop_name]
                    or prop_name in configured_properties
                )
                and valid_value
            ):
                calls_is_sbd_enabled.append(
                    mock.call(self.mock_service_manager)
                )
                if sbd_enabled:
                    calls_sbd_devices.append(mock.call())
                    if not sbd_devices and (
                        new_properties[prop_name]
                        not in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES
                    ):
                        calls_sbd_timeout.append(mock.call())

        self.mock_is_sbd_enabled.assert_has_calls(calls_is_sbd_enabled)
        self.mock_sbd_devices.assert_has_calls(calls_sbd_devices)
        self.mock_sbd_timeout.assert_has_calls(calls_sbd_timeout)
        self.assertEqual(
            self.mock_is_sbd_enabled.call_count, len(calls_is_sbd_enabled)
        )
        self.assertEqual(
            self.mock_sbd_devices.call_count, len(calls_sbd_devices)
        )
        self.assertEqual(
            self.mock_sbd_timeout.call_count, len(calls_sbd_timeout)
        )

    def test_no_properties_to_set_or_unset(self):
        self.assert_validate_set(
            [],
            {},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                )
            ],
        )

    def test_no_properties_to_set_or_unset_forced(self):
        self.assert_validate_set(
            [],
            {},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                )
            ],
            force=True,
        )

    def test_set_valid_properties_and_values(self):
        self.assert_validate_set([], FIXTURE_VALID_OPTIONS_DICT, [])

    def test_set_invalid_properties_and_values(self):
        self.assert_validate_set(
            [],
            FIXTURE_INVALID_OPTIONS_DICT,
            FIXTURE_ERROR_REPORTS,
            returncode=1,
        )

    def test_set_invalid_properties_and_values_forced(self):
        self.assert_validate_set(
            [],
            FIXTURE_INVALID_OPTIONS_DICT,
            warning_reports(FIXTURE_ERROR_REPORTS),
            force=True,
            returncode=1,
        )

    def test_unset_stonith_watchdog_timeout_sbd_disabled(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [prop_name], {prop_name: value}, []
                    )

    def test_set_stonith_watchdog_timeout_sbd_disabled(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    [],
                    {prop_name: "5"},
                    [
                        fixture.error(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                            reason="sbd_not_set_up",
                        )
                    ],
                )

    def test_set_ok_stonith_watchdog_timeout_sbd_enabled_without_devices(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in ["15", "15s"]:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [],
                        {prop_name: value},
                        [],
                        sbd_enabled=True,
                    )

    def test_set_small_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    [],
                    {prop_name: "9s"},
                    [
                        fixture.error(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                            force_code=reports.codes.FORCE,
                            cluster_sbd_watchdog_timeout=10,
                            entered_watchdog_timeout="9s",
                        )
                    ],
                    sbd_enabled=True,
                )

    def test_set_small_stonith_watchdog_timeout_sbd_enabled_without_devices_forced(
        self,
    ):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    [],
                    {prop_name: "9"},
                    [
                        fixture.warn(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                            cluster_sbd_watchdog_timeout=10,
                            entered_watchdog_timeout="9",
                        )
                    ],
                    sbd_enabled=True,
                    force=True,
                )

    def _set_invalid_value_stonith_watchdog_timeout(
        self, sbd_enabled=False, sbd_devices=False
    ):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in ["invalid", "10x"]:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [],
                        {prop_name: value},
                        [
                            fixture.error(
                                reports.codes.INVALID_OPTION_VALUE,
                                option_name=prop_name,
                                option_value=value,
                                allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                                cannot_be_empty=False,
                                forbidden_characters=None,
                            )
                        ],
                        sbd_enabled=sbd_enabled,
                        sbd_devices=sbd_devices,
                        valid_value=False,
                    )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=True, sbd_devices=False
        )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_enabled_with_devices(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=True, sbd_devices=True
        )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_disabled(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=False, sbd_devices=False
        )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [prop_name],
                        {prop_name: value},
                        [
                            fixture.error(
                                reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                                force_code=reports.codes.FORCE,
                                reason="sbd_set_up_without_devices",
                            )
                        ],
                        sbd_enabled=True,
                    )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_without_devices_forced(
        self,
    ):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [prop_name],
                        {prop_name: value},
                        [
                            fixture.warn(
                                reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                                reason="sbd_set_up_without_devices",
                            )
                        ],
                        force=True,
                        sbd_enabled=True,
                    )

    def test_set_stonith_watchdog_timeout_sbd_enabled_with_devices(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    [],
                    {prop_name: "15"},
                    [
                        fixture.error(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                            force_code=reports.codes.FORCE,
                            reason="sbd_set_up_with_devices",
                        )
                    ],
                    sbd_enabled=True,
                    sbd_devices=True,
                )

    def test_set_stonith_watchdog_timeout_sbd_enabled_with_devices_forced(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    [],
                    {prop_name: "15s"},
                    [
                        fixture.warn(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                            reason="sbd_set_up_with_devices",
                        )
                    ],
                    sbd_enabled=True,
                    sbd_devices=True,
                    force=True,
                )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_with_devices(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
                with self.subTest(property=prop_name, value=value):
                    self.assert_validate_set(
                        [prop_name],
                        {prop_name: value},
                        [],
                        sbd_enabled=True,
                        sbd_devices=True,
                    )

    def test_remove_not_configured_stonith_watchdog_timeout(self):
        for prop_name in STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
            with self.subTest(property=prop_name):
                self.assert_validate_set(
                    ["a", "b"],
                    {prop_name: ""},
                    [
                        fixture.error(
                            reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                            force_code=reports.codes.FORCE,
                            container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                            item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                            container_id="property-set-id",
                            item_list=[prop_name],
                        )
                    ],
                )

    def test_warn_about_disable_fencing(self):
        code = reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT_DUE_TO_PROPERTIES

        def assert_no_fencing_warning(property_map):
            self.assert_validate_set(
                [],
                property_map,
                [fixture.warn(code, property_map=property_map)],
            )

        for falsy_value in packemaker_values._BOOLEAN_FALSE:
            with self.subTest(value=falsy_value):
                assert_no_fencing_warning(
                    {
                        "stonith-enabled": falsy_value,
                        "fencing-enabled": falsy_value,
                    }
                )
            with self.subTest(value=falsy_value):
                assert_no_fencing_warning({"stonith-enabled": falsy_value})

            with self.subTest(value=falsy_value):
                assert_no_fencing_warning({"fencing-enabled": falsy_value})

    def test_remove_valid_configured_options(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"bool_param": "", "integer_param": ""},
            [],
        )

    def test_remove_valid_not_configured_options(self):
        self.assert_validate_set(
            [],
            {"bool_param": "", "integer_param": ""},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["bool_param", "integer_param"],
                )
            ],
        )

    def test_remove_valid_not_configured_options_forced(self):
        self.assert_validate_set(
            [],
            {"bool_param": "", "integer_param": ""},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["bool_param", "integer_param"],
                )
            ],
            force=True,
        )

    def test_remove_invalid_configured_options(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"a": "", "b": ""},
            [],
        )

    def test_remove_invalid_configured_options_forced(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"a": "", "b": ""},
            [],
            force=True,
        )

    def test_remove_invalid_not_configured_options(self):
        self.assert_validate_set(
            ["a", "b", "c"],
            {"x": "", "y": ""},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["x", "y"],
                ),
            ],
        )

    def test_remove_invalid_not_configured_options_forced(self):
        self.assert_validate_set(
            ["a", "b", "c"],
            {"x": "", "y": ""},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["x", "y"],
                ),
            ],
            force=True,
        )

    def test_remove_forbidden_options(self):
        self.assert_validate_set(
            FORBIDDEN_OPTIONS_LIST,
            dict.fromkeys(FORBIDDEN_OPTIONS_LIST, ""),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=FORBIDDEN_OPTIONS_LIST,
                    allowed=ALLOWED_PROPERTIES,
                    option_type="cluster property",
                    allowed_patterns=[],
                )
            ],
        )

    def test_remove_forbidden_options_forced(self):
        self.assert_validate_set(
            FORBIDDEN_OPTIONS_LIST,
            dict.fromkeys(FORBIDDEN_OPTIONS_LIST, ""),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=FORBIDDEN_OPTIONS_LIST,
                    allowed=ALLOWED_PROPERTIES,
                    option_type="cluster property",
                    allowed_patterns=[],
                )
            ],
            force=True,
        )


class TestGetClusterPropertySetElementLegacy(TestCase):
    def test_return_first_set(self):
        # pylint: disable=no-self-use
        cib = etree.fromstring(FIXTURE_TWO_PROPERTY_SETS)
        id_provider = IdProvider(cib)
        set_element = (
            lib_cluster_property.get_cluster_property_set_element_legacy(
                cib, id_provider
            )
        )
        assert_xml_equal(
            etree_to_str(cib.xpath('.//*[@id="first"]')[0]),
            etree_to_str(set_element),
        )

    def test_no_set_create_cib_bootstrap_options_set(self):
        # pylint: disable=no-self-use
        cib = etree.fromstring(FIXTURE_NO_SET)
        id_provider = IdProvider(cib)
        set_element = (
            lib_cluster_property.get_cluster_property_set_element_legacy(
                cib, id_provider
            )
        )
        assert_xml_equal(
            '<cluster_property_set id="cib-bootstrap-options"/>',
            etree_to_str(set_element),
        )



from typing import Optional

from pcs.common import file_type_codes
from pcs.common.reports.dto import ReportItemContextDto

_file_role_translation = {
    file_type_codes.BOOTH_CONFIG: "Booth configuration",
    file_type_codes.BOOTH_KEY: "Booth key",
    file_type_codes.COROSYNC_AUTHKEY: "Corosync authkey",
    file_type_codes.COROSYNC_CONF: "Corosync configuration",
    file_type_codes.COROSYNC_QDEVICE_NSSDB: "QDevice certificate database",
    file_type_codes.COROSYNC_QNETD_CA_CERT: "QNetd CA certificate",
    file_type_codes.COROSYNC_QNETD_NSSDB: "QNetd certificate database",
    file_type_codes.PCS_DR_CONFIG: "disaster-recovery configuration",
    file_type_codes.PACEMAKER_AUTHKEY: "Pacemaker authkey",
    file_type_codes.PCSD_ENVIRONMENT_CONFIG: "pcsd configuration",
    file_type_codes.PCSD_SSL_CERT: "pcsd SSL certificate",
    file_type_codes.PCSD_SSL_KEY: "pcsd SSL key",
    file_type_codes.PCS_CFGSYNC_CTL: "Config synchronization configuration",
    file_type_codes.PCS_KNOWN_HOSTS: "known-hosts",
    file_type_codes.PCS_SETTINGS_CONF: "pcs configuration",
}


def add_context_to_message(
    msg: str, context: Optional[ReportItemContextDto]
) -> str:
    if context:
        msg = f"{context.node}: {msg}"
    return msg


def format_file_role(role: file_type_codes.FileTypeCode) -> str:
    return _file_role_translation.get(role, role)



from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.reports.preprocessor import (
    get_duplicate_constraint_exists_preprocessor,
)
from pcs.common import reports
from pcs.common.pacemaker.constraint import CibConstraintsDto
from pcs.common.pacemaker.constraint.location import (
    CibConstraintLocationAttributesDto,
    CibConstraintLocationDto,
)
from pcs.lib.errors import LibraryError


class Preprocessor(TestCase):
    def setUp(self):
        self.mock_get_config = mock.Mock()
        self.mock_lib = mock.Mock(spec_set=["constraint"])
        self.mock_lib.constraint = mock.Mock(spec_set=["get_config"])
        self.mock_lib.constraint.get_config = self.mock_get_config

    @mock.patch("pcs.cli.reports.preprocessor.print_to_stderr")
    def test_cib_cache(self, mock_print_stderr):
        self.mock_get_config.return_value = CibConstraintsDto(
            [
                CibConstraintLocationDto(
                    "R1",
                    None,
                    None,
                    CibConstraintLocationAttributesDto(
                        "location1", "INFINITY", "node1", [], [], None
                    ),
                ),
                CibConstraintLocationDto(
                    "R2",
                    None,
                    None,
                    CibConstraintLocationAttributesDto(
                        "location2", "INFINITY", "node2", [], [], None
                    ),
                ),
            ]
        )

        preprocessor = get_duplicate_constraint_exists_preprocessor(
            self.mock_lib
        )
        preprocessor(
            reports.ReportItem.error(
                reports.messages.DuplicateConstraintsExist(["location1"])
            )
        )
        preprocessor(
            reports.ReportItem.error(
                reports.messages.DuplicateConstraintsExist(["location2"])
            )
        )

        stderr = dedent(
            """\
            Duplicate constraints:
              resource 'R1' prefers node 'node1' with score INFINITY (id: location1)
            Duplicate constraints:
              resource 'R2' prefers node 'node2' with score INFINITY (id: location2)
            """
        )
        stderr_calls = [mock.call(item) for item in stderr.splitlines()]
        mock_print_stderr.assert_has_calls(stderr_calls)
        self.assertEqual(mock_print_stderr.call_count, len(stderr_calls))
        self.mock_get_config.assert_called_once_with(evaluate_rules=False)

    @mock.patch("pcs.cli.reports.preprocessor.print_to_stderr")
    @mock.patch("pcs.cli.reports.output.print_to_stderr")
    def test_error_when_loading_constraints(
        self, mock_print_stderr_o, mock_print_stderr_p
    ):
        self.mock_get_config.side_effect = LibraryError(
            reports.ReportItem.error(reports.messages.CibLoadError("reason"))
        )
        preprocessor = get_duplicate_constraint_exists_preprocessor(
            self.mock_lib
        )
        preprocessor(
            reports.ReportItem.error(
                reports.messages.DuplicateConstraintsExist(["id2", "id1"])
            )
        )
        mock_print_stderr_o.assert_called_once_with("Error: unable to get cib")
        mock_print_stderr_p.assert_called_once_with(
            "Duplicate constraints: 'id1', 'id2'"
        )



import json
from dataclasses import replace
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs import resource
from pcs.cli.common.errors import CmdLineInputError
from pcs.common.interface import dto
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.reports import codes as report_codes
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)

from pcs_test.tools.misc import dict_to_modifiers

FIXTURE_EXPIRED_RULE = CibRuleExpressionDto(
    "my-meta-rule",
    CibRuleExpressionType.RULE,
    CibRuleInEffectStatus.EXPIRED,
    {"boolean-op": "and", "score": "INFINITY"},
    None,
    None,
    [
        CibRuleExpressionDto(
            "my-meta-rule-rsc",
            CibRuleExpressionType.RSC_EXPRESSION,
            CibRuleInEffectStatus.UNKNOWN,
            {
                "class": "ocf",
                "provider": "pacemaker",
                "type": "Dummy",
            },
            None,
            None,
            [],
            "resource ocf:pacemaker:Dummy",
        ),
    ],
    "resource ocf:pacemaker:Dummy",
)

FIXTURE_EXPIRED_RULE_NO_EXPIRE_CHECK = replace(
    FIXTURE_EXPIRED_RULE, in_effect=CibRuleInEffectStatus.UNKNOWN
)

FIXTURE_NVSET_EXPIRED_RULE = CibNvsetDto(
    "my-meta_attributes",
    {},
    FIXTURE_EXPIRED_RULE,
    [
        CibNvpairDto("my-id-pair1", "name1", "value1"),
        CibNvpairDto("my-id-pair2", "name2", "value2"),
    ],
)

FIXTURE_NVSET_EXPIRED_RULE_NO_EXPIRE_CHECK = replace(
    FIXTURE_NVSET_EXPIRED_RULE, rule=FIXTURE_EXPIRED_RULE_NO_EXPIRE_CHECK
)

FIXTURE_NVSET_NO_RULE = CibNvsetDto(
    "meta-plain",
    {"score": "123"},
    None,
    [CibNvpairDto("my-id-pair3", "name 1", "value 1")],
)

FIXTURE_INSTANCE_ATTRIBUTES = [
    CibNvsetDto(
        "instance",
        {},
        None,
        [CibNvpairDto("instance-pair", "inst", "ance")],
    ),
]
RULE_ARGV_DEPRECATED = (
    "Specifying a rule as multiple arguments is deprecated and might be removed "
    "in a future release, specify the rule as a single string instead"
)


class DefaultsBaseMixin:
    cli_command_name = ""
    lib_command_name = ""
    defaults_command = ""

    def setUp(self):
        self.lib = mock.Mock(spec_set=["cib_options"])
        self.cib_options = mock.Mock(spec_set=[self.lib_command_name])
        self.lib.cib_options = self.cib_options
        self.lib_command = getattr(self.cib_options, self.lib_command_name)
        self.cli_command = getattr(resource, self.cli_command_name)

    def _call_cmd(self, argv, modifiers=None):
        modifiers = modifiers or {}
        self.cli_command(self.lib, argv, dict_to_modifiers(modifiers))


@mock.patch("pcs.resource.print")
class DefaultsConfigMixin(DefaultsBaseMixin):
    empty_dto = CibDefaultsDto(instance_attributes=[], meta_attributes=[])
    dto_list = CibDefaultsDto(
        meta_attributes=[
            FIXTURE_NVSET_EXPIRED_RULE,
            FIXTURE_NVSET_NO_RULE,
        ],
        instance_attributes=FIXTURE_INSTANCE_ATTRIBUTES,
    )
    dto_list_expired_excluded = CibDefaultsDto(
        meta_attributes=[FIXTURE_NVSET_NO_RULE],
        instance_attributes=FIXTURE_INSTANCE_ATTRIBUTES,
    )
    dto_list_no_expire_check = CibDefaultsDto(
        meta_attributes=[
            FIXTURE_NVSET_EXPIRED_RULE_NO_EXPIRE_CHECK,
            FIXTURE_NVSET_NO_RULE,
        ],
        instance_attributes=FIXTURE_INSTANCE_ATTRIBUTES,
    )

    def test_no_args(self, mock_print):
        self.lib_command.return_value = self.empty_dto
        self._call_cmd([])
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_not_called()

    def test_usage(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_full(self, mock_print):
        self.lib_command.return_value = self.empty_dto
        self._call_cmd([], {"full": True})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_not_called()

    def test_full_with_no_text_format(self, mock_print):
        self.lib_command.return_value = self.empty_dto
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"full": True, "output-format": "json"})
        self.assertEqual(
            cm.exception.message,
            "option '--full' is not compatible with 'json' output format.",
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_no_expire_check(self, mock_print):
        self.lib_command.return_value = self.empty_dto
        self._call_cmd([], {"no-expire-check": True})
        self.lib_command.assert_called_once_with(False)
        mock_print.assert_not_called()

    def test_no_expire_check_and_include_expired(self, mock_print):
        self.lib_command.return_value = self.empty_dto
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True, "no-expire-check": True})
        self.assertEqual(
            cm.exception.message,
            "Only one of '--all', '--no-expire-check' can be used",
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_print_include_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"all": True})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            dedent(
                '''\
                Meta Attrs (expired): my-meta_attributes
                  name1=value1
                  name2=value2
                  Rule (expired): boolean-op=and score=INFINITY
                    Expression: resource ocf:pacemaker:Dummy
                Meta Attrs: meta-plain score=123
                  "name 1"="value 1"'''
            )
        )

    def test_print_exclude_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            dedent(
                '''\
                Meta Attrs: meta-plain score=123
                  "name 1"="value 1"'''
            )
        )

    def test_print_no_expire_check(self, mock_print):
        self.lib_command.return_value = self.dto_list_no_expire_check
        self._call_cmd([], {"no-expire-check": True})
        self.lib_command.assert_called_once_with(False)
        mock_print.assert_called_once_with(
            dedent(
                '''\
                Meta Attrs: my-meta_attributes
                  name1=value1
                  name2=value2
                  Rule: boolean-op=and score=INFINITY
                    Expression: resource ocf:pacemaker:Dummy
                Meta Attrs: meta-plain score=123
                  "name 1"="value 1"'''
            )
        )

    def test_print_full(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"all": True, "full": True})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            dedent(
                """\
                Meta Attrs (expired): my-meta_attributes
                  name1=value1 (id: my-id-pair1)
                  name2=value2 (id: my-id-pair2)
                  Rule (expired): boolean-op=and score=INFINITY (id: my-meta-rule)
                    Expression: resource ocf:pacemaker:Dummy (id: my-meta-rule-rsc)
                Meta Attrs: meta-plain score=123
                  "name 1"="value 1" (id: my-id-pair3)"""
            )
        )

    def test_print_format_cmd_exclude_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"output-format": "cmd"})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            dedent(
                f"""\
                pcs -- resource {self.defaults_command} set create id=meta-plain score=123 \\
                  meta 'name 1=value 1'"""
            )
        )

    def _fixture_commands_all(self):
        return dedent(
            f"""\
                pcs -- resource {self.defaults_command} set create id=my-meta_attributes \\
                  meta name1=value1 name2=value2 \\
                  rule 'resource ocf:pacemaker:Dummy';
                pcs -- resource {self.defaults_command} set create id=meta-plain score=123 \\
                  meta 'name 1=value 1'"""
        )

    def test_print_format_cmd_include_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"all": True, "output-format": "cmd"})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(self._fixture_commands_all())

    def test_print_format_cmd_no_expire_check(self, mock_print):
        self.lib_command.return_value = self.dto_list_no_expire_check
        self._call_cmd([], {"no-expire-check": True, "output-format": "cmd"})
        self.lib_command.assert_called_once_with(False)
        mock_print.assert_called_once_with(self._fixture_commands_all())

    def test_print_format_json_exclude_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"output-format": "json"})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(self.dto_list_expired_excluded))
        )

    def test_print_format_json_include_expired(self, mock_print):
        self.lib_command.return_value = self.dto_list
        self._call_cmd([], {"all": True, "output-format": "json"})
        self.lib_command.assert_called_once_with(True)
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(self.dto_list))
        )


class RscDefaultsConfig(DefaultsConfigMixin, TestCase):
    cli_command_name = "resource_defaults_config_cmd"
    lib_command_name = "resource_defaults_config"
    defaults_command = "defaults"


class OpDefaultsConfig(DefaultsConfigMixin, TestCase):
    cli_command_name = "resource_op_defaults_config_cmd"
    lib_command_name = "operation_defaults_config"
    defaults_command = "op defaults"


class DefaultsSetCreateMixin(DefaultsBaseMixin):
    def test_no_args(self):
        self._call_cmd([])
        self.lib_command.assert_called_once_with(
            {}, {}, nvset_rule=None, force_flags=set()
        )

    def test_no_values(self):
        self._call_cmd(["meta", "rule"])
        self.lib_command.assert_called_once_with(
            {}, {}, nvset_rule=None, force_flags=set()
        )

    def test_bad_options_or_keyword(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["aaa"])
        self.assertEqual(
            cm.exception.message,
            "missing value of 'aaa' option",
        )
        self.lib_command.assert_not_called()

    def test_bad_values(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["meta", "aaa"])
        self.assertEqual(
            cm.exception.message,
            "missing value of 'aaa' option",
        )
        self.lib_command.assert_not_called()

    def test_options(self):
        self._call_cmd(["id=custom-id", "score=10"])
        self.lib_command.assert_called_once_with(
            {},
            {"id": "custom-id", "score": "10"},
            nvset_rule=None,
            force_flags=set(),
        )

    def test_nvpairs(self):
        self._call_cmd(["meta", "name1=value1", "name2=value2"])
        self.lib_command.assert_called_once_with(
            {"name1": "value1", "name2": "value2"},
            {},
            nvset_rule=None,
            force_flags=set(),
        )

    def test_rule(self):
        self._call_cmd(["rule", "resource dummy or op monitor"])
        self.lib_command.assert_called_once_with(
            {},
            {},
            nvset_rule="resource dummy or op monitor",
            force_flags=set(),
        )

    @mock.patch("pcs.cli.common.parse_args.deprecation_warning")
    def test_rule_deprecated_form(self, mock_dw):
        self._call_cmd(["rule", "resource", "dummy", "or", "op", "monitor"])
        self.lib_command.assert_called_once_with(
            {},
            {},
            nvset_rule="resource dummy or op monitor",
            force_flags=set(),
        )
        mock_dw.assert_called_once_with(RULE_ARGV_DEPRECATED)

    def test_force(self):
        self._call_cmd([], {"force": True})
        self.lib_command.assert_called_once_with(
            {}, {}, nvset_rule=None, force_flags={report_codes.FORCE}
        )

    def test_all(self):
        self._call_cmd(
            [
                "id=custom-id",
                "score=10",
                "meta",
                "name1=value1",
                "name2=value2",
                "rule",
                "resource dummy or op monitor",
            ],
            {"force": True},
        )
        self.lib_command.assert_called_once_with(
            {"name1": "value1", "name2": "value2"},
            {"id": "custom-id", "score": "10"},
            nvset_rule="resource dummy or op monitor",
            force_flags={report_codes.FORCE},
        )


class RscDefaultsSetCreate(DefaultsSetCreateMixin, TestCase):
    cli_command_name = "resource_defaults_set_create_cmd"
    lib_command_name = "resource_defaults_create"


class OpDefaultsSetCreate(DefaultsSetCreateMixin, TestCase):
    cli_command_name = "resource_op_defaults_set_create_cmd"
    lib_command_name = "operation_defaults_create"


class DefaultsSetRemoveMixin(DefaultsBaseMixin):
    def test_no_args(self):
        self._call_cmd([])
        self.lib_command.assert_called_once_with([])

    def test_some_args(self):
        self._call_cmd(["set1", "set2"])
        self.lib_command.assert_called_once_with(["set1", "set2"])


class RscDefaultsSetRemove(DefaultsSetRemoveMixin, TestCase):
    cli_command_name = "resource_defaults_set_remove_cmd"
    lib_command_name = "resource_defaults_remove"


class OpDefaultsSetRemove(DefaultsSetRemoveMixin, TestCase):
    cli_command_name = "resource_op_defaults_set_remove_cmd"
    lib_command_name = "operation_defaults_remove"


class DefaultsSetUpdateMixin(DefaultsBaseMixin):
    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()

    def test_no_meta(self):
        self._call_cmd(["nvset-id"])
        self.lib_command.assert_called_once_with("nvset-id", {})

    def test_no_meta_values(self):
        self._call_cmd(["nvset-id", "meta"])
        self.lib_command.assert_called_once_with("nvset-id", {})

    def test_meta_values(self):
        self._call_cmd(["nvset-id", "meta", "a=b", "c=d"])
        self.lib_command.assert_called_once_with(
            "nvset-id", {"a": "b", "c": "d"}
        )


class RscDefaultsSetUpdate(DefaultsSetUpdateMixin, TestCase):
    cli_command_name = "resource_defaults_set_update_cmd"
    lib_command_name = "resource_defaults_update"


class OpDefaultsSetUpdate(DefaultsSetUpdateMixin, TestCase):
    cli_command_name = "resource_op_defaults_set_update_cmd"
    lib_command_name = "operation_defaults_update"


class DefaultsUpdateMixin(DefaultsBaseMixin):
    def test_no_args(self):
        self._call_cmd([])
        self.lib_command.assert_called_once_with(None, {})

    def test_args(self):
        self._call_cmd(["a=b", "c="])
        self.lib_command.assert_called_once_with(None, {"a": "b", "c": ""})


class RscDefaultsUpdate(DefaultsUpdateMixin, TestCase):
    cli_command_name = "resource_defaults_update_cmd"
    lib_command_name = "resource_defaults_update"


class OpDefaultsUpdate(DefaultsUpdateMixin, TestCase):
    cli_command_name = "resource_op_defaults_update_cmd"
    lib_command_name = "operation_defaults_update"



from pcs.common.pacemaker.node import CibNodeDto, CibNodeListDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import CibRuleExpressionType, CibRuleInEffectStatus
from pcs.lib.cib.rule.in_effect import RuleInEffectEval

FIXTURE_NODES_CONFIG_XML = """
    <nodes>
        <node id="1" uname="node1" description="node1 desc" score="100"
            type="member"
        >
            <instance_attributes id="nodes-1" score="50">
                <nvpair id="nodes-1-a" name="a" value="1"/>
                <nvpair id="nodes-1-b" name="b" value="2"/>
                <rule id="nodes-1-rule" score="INFINITY">
                    <date_expression id="nodes-1-rule-de" operation="lt"
                        end="2000-01-01"
                    />
                </rule>
            </instance_attributes>
            <instance_attributes id="nodes-1-2">
                <nvpair id="nodes-1-2-c" name="c" value="3"/>
                <rule id="nodes-1-2-rule" boolean-op="and" score="INFINITY">
                    <date_expression id="nodes-1-2-rule-de" operation="gt"
                        start="2000-01-01"
                    />
                </rule>
            </instance_attributes>
            <utilization id="nodes-1-utilization" score="50">
                <nvpair id="nodes-1-utilization-cpu" name="cpu" value="4"/>
                <nvpair id="nodes-1-utilization-ram" name="ram" value="32"/>
            </utilization>
        </node>
        <node id="2" uname="node2" score="200" type="remote">
            <instance_attributes id="nodes-2">
                <nvpair id="nodes-2-a" name="a" value="1"/>
                <nvpair id="nodes-2-b" name="b" value="2"/>
            </instance_attributes>
            <utilization id="nodes-2-utilization">
                <nvpair id="nodes-2-utilization-cpu" name="cpu" value="8"/>
                <nvpair id="nodes-2-utilization-ram" name="ram" value="64"/>
                <rule id="nodes-2-utilization-rule" boolean-op="and"
                    score="INFINITY"
                >
                    <date_expression id="nodes-2-utilization-rule-de"
                        operation="gt" start="2000-01-01"/>
                </rule>
            </utilization>
        </node>
        <node id="3" uname="node3"/>
    </nodes>
"""


def get_nodes_dto(rule_eval: RuleInEffectEval) -> CibNodeListDto:
    return CibNodeListDto(
        nodes=[
            CibNodeDto(
                id="1",
                uname="node1",
                description="node1 desc",
                score="100",
                type="member",
                instance_attributes=[
                    CibNvsetDto(
                        id="nodes-1",
                        options={"score": "50"},
                        rule=CibRuleExpressionDto(
                            id="nodes-1-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status("nodes-1-rule"),
                            options={"score": "INFINITY"},
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="nodes-1-rule-de",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=CibRuleInEffectStatus.UNKNOWN,
                                    options={
                                        "operation": "lt",
                                        "end": "2000-01-01",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date lt 2000-01-01",
                                )
                            ],
                            as_string="date lt 2000-01-01",
                        ),
                        nvpairs=[
                            CibNvpairDto(id="nodes-1-a", name="a", value="1"),
                            CibNvpairDto(id="nodes-1-b", name="b", value="2"),
                        ],
                    ),
                    CibNvsetDto(
                        id="nodes-1-2",
                        options={},
                        rule=CibRuleExpressionDto(
                            id="nodes-1-2-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status(
                                "nodes-1-2-rule"
                            ),
                            options={"boolean-op": "and", "score": "INFINITY"},
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="nodes-1-2-rule-de",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=CibRuleInEffectStatus.UNKNOWN,
                                    options={
                                        "operation": "gt",
                                        "start": "2000-01-01",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date gt 2000-01-01",
                                )
                            ],
                            as_string="date gt 2000-01-01",
                        ),
                        nvpairs=[
                            CibNvpairDto(id="nodes-1-2-c", name="c", value="3"),
                        ],
                    ),
                ],
                utilization=[
                    CibNvsetDto(
                        id="nodes-1-utilization",
                        options={"score": "50"},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(
                                id="nodes-1-utilization-cpu",
                                name="cpu",
                                value="4",
                            ),
                            CibNvpairDto(
                                id="nodes-1-utilization-ram",
                                name="ram",
                                value="32",
                            ),
                        ],
                    )
                ],
            ),
            CibNodeDto(
                id="2",
                uname="node2",
                description=None,
                score="200",
                type="remote",
                instance_attributes=[
                    CibNvsetDto(
                        id="nodes-2",
                        options={},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(id="nodes-2-a", name="a", value="1"),
                            CibNvpairDto(id="nodes-2-b", name="b", value="2"),
                        ],
                    )
                ],
                utilization=[
                    CibNvsetDto(
                        id="nodes-2-utilization",
                        options={},
                        rule=CibRuleExpressionDto(
                            id="nodes-2-utilization-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status(
                                "nodes-2-utilization-rule"
                            ),
                            options={"boolean-op": "and", "score": "INFINITY"},
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="nodes-2-utilization-rule-de",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=CibRuleInEffectStatus.UNKNOWN,
                                    options={
                                        "operation": "gt",
                                        "start": "2000-01-01",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date gt 2000-01-01",
                                )
                            ],
                            as_string="date gt 2000-01-01",
                        ),
                        nvpairs=[
                            CibNvpairDto(
                                id="nodes-2-utilization-cpu",
                                name="cpu",
                                value="8",
                            ),
                            CibNvpairDto(
                                id="nodes-2-utilization-ram",
                                name="ram",
                                value="64",
                            ),
                        ],
                    )
                ],
            ),
            CibNodeDto(
                id="3",
                uname="node3",
                description=None,
                score=None,
                type=None,
                instance_attributes=[],
                utilization=[],
            ),
        ]
    )



from unittest import (
    TestCase,
    mock,
)

from pcs.common import capabilities

from pcs_test.tools.misc import get_test_resource as rc


@mock.patch("pcs.settings.pcs_capabilities", rc("capabilities.xml"))
class Capabilities(TestCase):
    def test_get_definition(self):
        self.assertEqual(
            capabilities.get_capabilities_definition(),
            [
                capabilities.Capability(
                    code="test.in-pcs",
                    in_pcs=True,
                    in_pcsd=False,
                    description="This capability is available in pcs.",
                ),
                capabilities.Capability(
                    code="test.in-pcsd",
                    in_pcs=False,
                    in_pcsd=True,
                    description="This capability is available in pcsd.",
                ),
                capabilities.Capability(
                    code="test.both",
                    in_pcs=True,
                    in_pcsd=True,
                    description=(
                        "This capability is available in both pcs and pcsd."
                    ),
                ),
                capabilities.Capability(
                    code="test.empty-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
                capabilities.Capability(
                    code="test.no-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
            ],
        )

    def test_get_pcs(self):
        self.assertEqual(
            capabilities.get_pcs_capabilities(),
            [
                capabilities.Capability(
                    code="test.in-pcs",
                    in_pcs=True,
                    in_pcsd=False,
                    description="This capability is available in pcs.",
                ),
                capabilities.Capability(
                    code="test.both",
                    in_pcs=True,
                    in_pcsd=True,
                    description=(
                        "This capability is available in both pcs and pcsd."
                    ),
                ),
                capabilities.Capability(
                    code="test.empty-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
                capabilities.Capability(
                    code="test.no-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
            ],
        )

    def test_get_pcsd(self):
        self.assertEqual(
            capabilities.get_pcsd_capabilities(),
            [
                capabilities.Capability(
                    code="test.in-pcsd",
                    in_pcs=False,
                    in_pcsd=True,
                    description="This capability is available in pcsd.",
                ),
                capabilities.Capability(
                    code="test.both",
                    in_pcs=True,
                    in_pcsd=True,
                    description=(
                        "This capability is available in both pcs and pcsd."
                    ),
                ),
                capabilities.Capability(
                    code="test.empty-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
                capabilities.Capability(
                    code="test.no-description",
                    in_pcs=True,
                    in_pcsd=True,
                    description="",
                ),
            ],
        )



from typing import Final

TAG_ACL_GROUP: Final = "acl_group"
TAG_ACL_PERMISSION: Final = "acl_permission"
TAG_ACL_ROLE: Final = "acl_role"
TAG_ACL_TARGET: Final = "acl_target"
TAG_ALERT: Final = "alert"
TAG_CONSTRAINT_COLOCATION: Final = "rsc_colocation"
TAG_CONSTRAINT_LOCATION: Final = "rsc_location"
TAG_CONSTRAINT_ORDER: Final = "rsc_order"
TAG_CONSTRAINT_TICKET: Final = "rsc_ticket"
TAG_CRM_CONFIG: Final = "crm_config"
TAG_FENCING_LEVEL: Final = "fencing-level"
TAG_NODE: Final = "node"
TAG_OBJREF: Final = "obj_ref"
TAG_RECIPIENT: Final = "recipient"
TAG_RESOURCE_BUNDLE: Final = "bundle"
TAG_RESOURCE_CLONE: Final = "clone"
TAG_RESOURCE_GROUP: Final = "group"
TAG_RESOURCE_MASTER: Final = "master"
TAG_RESOURCE_PRIMITIVE: Final = "primitive"
TAG_RESOURCE_REF: Final = "resource_ref"
TAG_RESOURCE_SET: Final = "resource_set"
TAG_ROLE: Final = "role"
TAG_RULE: Final = "rule"
TAG_TAG: Final = "tag"

TAG_LIST_CONSTRAINABLE: Final = frozenset(
    (
        TAG_RESOURCE_BUNDLE,
        TAG_RESOURCE_CLONE,
        TAG_RESOURCE_GROUP,
        TAG_RESOURCE_MASTER,
        TAG_RESOURCE_PRIMITIVE,
        # Not yet supported. What needs to be done:
        # * move all constraint code to library
        # * add support for using tags in constraints
        # * write tests to verify creating constraints with tags
        # * write tests verifying displaying constraints with tags
        # TAG_TAG,
    )
)
TAG_LIST_CONSTRAINT = frozenset(
    (
        TAG_CONSTRAINT_COLOCATION,
        TAG_CONSTRAINT_LOCATION,
        TAG_CONSTRAINT_ORDER,
        TAG_CONSTRAINT_TICKET,
    )
)
TAG_LIST_RESOURCE: Final = frozenset(
    (
        TAG_RESOURCE_BUNDLE,
        TAG_RESOURCE_CLONE,
        TAG_RESOURCE_GROUP,
        TAG_RESOURCE_MASTER,
        TAG_RESOURCE_PRIMITIVE,
    )
)
TAG_LIST_RESOURCE_MULTIINSTANCE: Final = frozenset(
    (
        TAG_RESOURCE_BUNDLE,
        TAG_RESOURCE_CLONE,
        TAG_RESOURCE_MASTER,
    )
)
